/*
 * PalmOSGUI.c
 * The GUI for the Palm. This source is 68k, unlike all the other files
 * which are compiled for ARM.
 */

#include <PalmOS.h>
#include <Bitmap.h>
#include <MemoryMgr.h>
#include <DataMgr.h>
#include <SoundMgr.h>
#include <PceNativeCall.h>

#include "PalmOSGUI.h"
#include "Armlet68k.h"
#include "FrodoResources.h"

#define DEBUG_GUI

UInt32 appSignature  = makeType('F','r','o','d');
DmResType armletType = makeType('A','r','m','l');
DmResType romType    = makeType('R','O','M','z');
DmResType tblType    = makeType('T','a','b','l');
DmResType strType    = makeType('S','t','r','g');
DmResType prefsType  = makeType('P','r','e','f');

UInt8 *armlet=NULL;
ArmletData *armletData=NULL;
#define SOUND_BUFSIZE 4096
#define FRAME_SAMPLES 882


UInt16 armletBuf[FRAME_SAMPLES]; // 1 frame of samples
struct SoundStruct {
	UInt16 soundBuf[SOUND_BUFSIZE]; // Ringbuffer for sound
	UInt16 sampleOffset, samplesQueued;
} soundStruct;
UInt8 canSound=0, can16=0, is320=0;
UInt32 screenW=0, screenH=0;
Int32 penX, penY;
Boolean penDown=false;
struct PrefsState *prefs;

struct {
	UInt16 numEntries;
	RGBColorType pal[16];
} colorTable =
{ 16,
{
	{ 0,   0,   0,   0},
	{ 1, 255, 255, 255},
	{ 2, 116,  67,  53},
	{ 3, 124, 172, 186},
	{ 4, 123,  72, 144},
	{ 5, 100, 151,  79},
	{ 6,  64,  50, 133},
	{ 7, 191, 205, 122},
	{ 8, 123,  91,  47},
	{ 9,  79,  69,   0},
	{10, 163, 114, 101},
	{11,  80,  80,  80},
	{12, 120, 120, 120},
	{13, 164, 215, 142},
	{14, 120, 106, 189},
	{15, 159, 159, 159}
}
};

#define swap32(x) (((((UInt32)(x))&0x000000ffL)<<24) | \
                   ((((UInt32)(x))&0x0000ff00L)<< 8) | \
                   ((((UInt32)(x))&0x00ff0000L)>> 8) | \
                   ((((UInt32)(x))&0xff000000L)>>24))
#define swap16(x) ((((UInt16)(x))&0x00ff)<<8) | \
                   (((UInt16)(x))&0xff00)>>8))

UInt32 dataSize=sizeof(ArmletData);
UInt8 *screenBits=NULL, *bmpBits=NULL;
BitmapType *bmp=NULL;
BitmapTypeV3 *bmpv3=NULL;
SndStreamRef sndStream=NULL;
char msg[101];

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt32 error=0;

	switch (cmd) {
		case sysAppLaunchCmdNormalLaunch:
			error = Launch();

			FreeGraphicsSound();
			FreeArmletData();
			FtrPtrFree(appSignature, 0);
			break;
		case sysAppLaunchCmdFind:
			//error = Search((FindParamsPtr) cmdPBP);
			break;
	}
	if (error) return error;
	return 0;
}

UInt32 Launch()
{
	UInt32 error=0;
	
	if (error = CheckPrerequisites())
		return error;
	if (error=CreateArmlet())
		return error;
	if (error=CreateArmletData()) {
		sprintf(msg, "Error in CreateArmletData (0x%lx)", error);
		SysFatalAlert(msg);
		return error;
	}
	if (error=InitGraphicsSound())
		return error;
	//SysFatalAlert("2");

	error = RunArmlet();

	switch (error) {
		case ERR_NORMAL_EXIT: error=0; break;
		case ERR_NEW_FRODO:
			SysFatalAlert("Out of memory creating Frodo"); break;
		case ERR_NEW_C64:
			SysFatalAlert("Out of memory creating C64"); break;
		case ERR_NEW_CPU:
			SysFatalAlert("Out of memory creating CPU"); break;
		case ERR_NEW_JOB1541:
			SysFatalAlert("Out of memory creating 1541Job"); break;
		case ERR_NEW_CPU1541:
			SysFatalAlert("Out of memory creating 1541CPU"); break;
		case ERR_NEW_VIC:
			SysFatalAlert("Out of memory creating VIC"); break;
		case ERR_NEW_SID:
			SysFatalAlert("Out of memory creating SID"); break;
		case ERR_NEW_CIA1:
			SysFatalAlert("Out of memory creating CIA1"); break;
		case ERR_NEW_CIA2:
			SysFatalAlert("Out of memory creating CIA2"); break;
		case ERR_NEW_IEC:
			SysFatalAlert("Out of memory creating IEC"); break;
		case ERR_NEW_REU:
			SysFatalAlert("Out of memory creating REU"); break;
		default:
			sprintf(msg, "Unknow error %ld", error);
			SysFatalAlert(msg);
	}

	return error;
}

UInt32 CheckPrerequisites()
{
	UInt32 w,h,d, error=0;
	Boolean c;

	// OS5+?
	error = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &d);
	if (error || d < sysMakeROMVersion(5,0,0,0,0)) {
		sprintf(msg, "You need PalmOS 5.0 or greater! (You have 0x%lx)",
			d);
		SysFatalAlert(msg);
		return 2;
	}
	// ARM processor?
	error = FtrGet(sysFileCSystem, sysFtrNumProcessorID, &d);
	if (error || !sysFtrNumProcessorIsARM(d)) {
		sprintf(msg, "You need an ARM processor to run Frodo (Your processor is 0x%lx)", d);
		SysFatalAlert(msg);
		return 7;
	}
	// Hires support?
	error = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &d);
	if (error || d < 4) {
		sprintf(msg, "You need high-res support "\
			"(Window Manager 4+). (You have 0x%lx)", d);
		SysFatalAlert(msg);
		return 3;
	}
	// Hires?
	WinScreenGetAttribute(winScreenDensity, &d);
	if (d < kDensityDouble) {
		sprintf(msg, "Your device doesn't have a screen with "\
			"320x320 (or more) pixels. (Your density is 0x%lx)", d);
		SysFatalAlert(msg);
		return 4;
	}
	if (d == kDensityDouble) is320 = 1;

	// Wider than 320?
	error=WinScreenGetAttribute(winScreenWidth, &screenW);
	if (error) {
		SysFatalAlert("Your Palm refuses to divulge its screen width!");
		return 9;
	}
	if (screenW<320) {
		sprintf(msg, "Your Palm said it was hires, but in fact it only"\
			" has a screen width of %ld (need 320).", screenW);
		SysFatalAlert(msg);
		return 10;
	}
	// Higher than 272?
	error=WinScreenGetAttribute(winScreenHeight, &screenH);
	if (error) {
		SysFatalAlert("Your Palm refuses to report its screen height!");
		return 11;
	}
	// Actually, right now I'm demanding 320 because I'm lazy and can't
	// bother to turn of the on-screen keyboard if there isn't room for it
	if (screenH<320) {
		sprintf(msg, "The screen size %ldx%ld isn't tall enough. "\
			"(Frodo needs a minimum of 320x320)", screenW, screenH);
		SysFatalAlert(msg);
		return 12;
	}

	// Some optional features
	// 16 color mode?
	c = true;
	error = WinScreenMode(winScreenModeGetSupportedDepths, &w, &h, &d, &c);
	if (!error && (d & 8)) can16 = 1;

	// Sampled sound?
	canSound=1;
	/*
	// Does NOT WORK AT ALL on 5.0 Tungsten
	error = FtrGet(sysFileCSoundMgr, sndFtrIDVersion, &d);
	if (error) {
		SysFatalAlert("Your device isn't capable of sampled sound. Sound is off.");
	} else {
		canSound = 1;
		soundStruct.sampleOffset=soundStruct.samplesQueued=0;
	}
	*/
	return 0;
}

#define MAXCHUNKS 5
UInt32 CreateArmlet()
{
	MemHandle armletPart[MAXCHUNKS];
	MemPtr p;
	UInt32 armletSize, offs;
	UInt8 nParts, i;
	UInt32 error;

	// Reassemble armlet from resources and stuff it in feature memory
	// Armlet is stored as Arml IDX_ARMLET, +1, +2, +3, ...
	armletSize=0;
	for (i=0; i<MAXCHUNKS; i++) {
		armletPart[i] = DmGetResource(armletType, IDX_ARMLET+i);
		if (!armletPart[i]) break;
		armletSize+=MemHandleSize(armletPart[i]);
	}
	if (i>=MAXCHUNKS) {
		sprintf(msg, "Too many chunks in armlet (%d)", i);
		SysFatalAlert(msg);
		for (i=0; i<MAXCHUNKS; i++) DmReleaseResource(armletPart[i]);
		return 1;
	}
	if (i==0) return DmGetLastErr();
	nParts=i;
	
	// Free feature 0 just in case we exited without freeing it last time
	FtrPtrFree(appSignature, 0);
	if (error = FtrPtrNew(appSignature, 0, armletSize, (void **) &armlet)) {
		SysFatalAlert("Couldn't create feature");
		for (i=0; i<nParts; i++) DmReleaseResource(armletPart[i]);
		return error;
	}

	offs=0;
	// Write the parts to feature memory
	for (i=0; i<nParts; i++) {
		p = MemHandleLock(armletPart[i]);
		error=DmWrite(armlet, offs, p, MemHandleSize(armletPart[i]));
		MemHandleUnlock(armletPart[i]);
		if (error) {
			sprintf(msg, "Couldn't write feature (part %d)", i);
			SysFatalAlert(msg);
			for (i=0; i<nParts; i++)
				DmReleaseResource(armletPart[i]);
			return error;
		}
		offs += MemHandleSize(armletPart[i]);
	}
	for (i=0; i<nParts; i++) DmReleaseResource(armletPart[i]);
	return 0;
}
	
// Allocate ArmletData and fill it in (mind the endianness!)
UInt32 CreateArmletData()
{
	UInt32 error, x;
	UInt8 i;
	DmOpenRef db;
	MemHandle h;

	armletData = (ArmletData *) MemPtrNew(sizeof(ArmletData));
	if (!armletData) return memErrNotEnoughSpace;
	MemSet(armletData, sizeof(ArmletData), 0x0);
	
	// Callbacks
	x = (UInt32) VBlank68k;
	armletData->vBlank    = (void *) swap32(x);
	x = (UInt32) ShowRequester68k;
	armletData->requester = (void *) swap32(x);
	x = (UInt32) Malloc68k;
	armletData->malloc    = (void *) swap32(x);
	x = (UInt32) Free68k;
	armletData->free      = (void *) swap32(x);
										
	// RAM & ROM
	armletData->ram1 = (UInt8 *) NewSwapped(0x8000); // 32 kB
	if (!armletData->ram1) return memErrNotEnoughSpace;
	armletData->ram2 = (UInt8 *) NewSwapped(0x8000); // 32 kB
	if (!armletData->ram2) return memErrNotEnoughSpace;
	armletData->kernalROM = CopyRes(romType, IDX_KERNAL, "KernalROM");
	if (!armletData->kernalROM) return memErrNotEnoughSpace;
	armletData->basicROM = GetRes(romType, IDX_BASIC, "BasicROM");
	if (!armletData->basicROM) return memErrNotEnoughSpace;
	armletData->charROM = GetRes(romType, IDX_CHAR, "CharROM");
	if (!armletData->charROM) return memErrNotEnoughSpace;
	armletData->color = (UInt8 *) NewSwapped(0x0400); // Color RAM
	if (!armletData->color) return memErrNotEnoughSpace;
	armletData->ram1541 = (UInt8 *) NewSwapped(0x0800); // Diskdrive RAM
	if (!armletData->ram1541) return memErrNotEnoughSpace;
	armletData->rom1541 = CopyRes(romType, IDX_1541, "1541ROM");
	if (!armletData->rom1541) return memErrNotEnoughSpace;

	// Various tables and precalculations
	armletData->expTable = GetRes(tblType, IDX_EXPTABLE, "ExpTable");
	if (!armletData->expTable) return memErrNotEnoughSpace;
	armletData->multiExpTable=GetRes(tblType, IDX_MULTIEXPTABLE,
		"MultiExpTable");
	if (!armletData->multiExpTable) return memErrNotEnoughSpace;
	armletData->textColorTable=GetRes(tblType, IDX_TEXTCOLORTABLE,
		"Text color lookup table");
	if (!armletData->textColorTable) return memErrNotEnoughSpace;
	armletData->triTable=GetRes(tblType, IDX_TRITABLE, "TriTable");
	if (!armletData->triTable) return memErrNotEnoughSpace;
	armletData->triSawTable=GetRes(tblType, IDX_TRISAWTABLE, "TriSawTable");
	if (!armletData->triSawTable) return memErrNotEnoughSpace;
	armletData->triRectTable=GetRes(tblType, IDX_TRIRECTTABLE,
		"TriRectTable");
	if (!armletData->triRectTable) return memErrNotEnoughSpace;
	armletData->sawRectTable=GetRes(tblType, IDX_SAWRECTTABLE,
		"SawRectTable");
	if (!armletData->sawRectTable) return memErrNotEnoughSpace;
	armletData->triSawRectTable=GetRes(tblType, IDX_TRISAWRECTTABLE,
		"TriSawRectTable");
	if (!armletData->triSawRectTable) return memErrNotEnoughSpace;
	armletData->egTable=GetRes(tblType, IDX_EGTABLE, "EGTable");
	if (!armletData->egTable) return memErrNotEnoughSpace;
	armletData->egdrShift=GetRes(tblType, IDX_EGDRSHIFT, "EGDRShift");
	if (!armletData->egdrShift) return memErrNotEnoughSpace;
	armletData->sampleTab=GetRes(tblType, IDX_SAMPLETAB, "SampleTab");
	if (!armletData->sampleTab) return memErrNotEnoughSpace;
	armletData->sintab=GetRes(tblType, IDX_SINTAB, "SinTab");
	if (!armletData->sintab) return memErrNotEnoughSpace;
	armletData->resonanceLP=GetRes(tblType, IDX_RESONANCELP, "ResonanceLP");
	if (!armletData->resonanceLP) return memErrNotEnoughSpace;
	armletData->resonanceHP=GetRes(tblType, IDX_RESONANCEHP, "ResonanceLP");
	if (!armletData->resonanceHP) return memErrNotEnoughSpace;

	for (i=0; i<12; i++) {
		armletData->errors1541[i]=GetRes(strType, IDX_ERRORS1541+i,
			"Errors1541");
		if (!armletData->errors1541[i])
			return memErrNotEnoughSpace;
	}
	// Set up prefs
	armletData->setPrefs=1; // Normally we want the Armlet to set it's prefs
	armletData->prefs=CopyRes2(prefsType, IDX_PREFS, "Preferences", 1);
	if (!armletData->prefs) {
		if (!(db = DmNextOpenResDatabase(NULL))) {
			SysFatalAlert("No open database!?");
			return DmGetLastErr();
		}
		if (!(h = DmNewResource(db, prefsType, IDX_PREFS,
				sizeof(struct PrefsState)))) {
			error=DmGetLastErr();
			SysFatalAlert("Couldn't create preferences");
			return error;
		}
		DmReleaseResource(h);
		armletData->prefs=CopyRes(prefsType, IDX_PREFS, "Preferences");
		if (!armletData->prefs) {
			return memErrNotEnoughSpace;
		}
		MemSet((MemPtr)swap32(armletData->prefs), sizeof(struct PrefsState), 0x0);
	}
	// store a normal-endian copy in prefs
	prefs=(struct PrefsState *) swap32(armletData->prefs);
	if (IsNukedPrefs(prefs)) {
		// Tell armlet to give us the default prefs
		armletData->getPrefs=1;
		armletData->setPrefs=0;
	}
#ifdef DEBUG_GUI
	prefs->Sound=1;
	prefs->AdaptiveFrameSkip=1;
	prefs->LimitSpeed=1;
#endif
	
	if (!canSound) prefs->Sound=0;

	return 0;
}

FreeArmletData()
{
	int i;
	
	if (!armletData) return;

	// Free up everything...
	// Things that where allocated by MemPtrNew or CopyRes
	// should get passed to MemPtrFree,
	// things that where created by GetResource should have their handles
	// unlocked.

	FreeSwapped(armletData->ram1);
	FreeSwapped(armletData->ram2);
	UncopyRes(armletData->kernalROM);
	UngetRes(armletData->basicROM);
	UngetRes(armletData->charROM);
	FreeSwapped(armletData->color);
	FreeSwapped(armletData->ram1541);
	UncopyRes(armletData->rom1541);
	UngetRes(armletData->expTable);
	UngetRes(armletData->multiExpTable);
	UngetRes(armletData->textColorTable);
	UngetRes(armletData->triTable);
	UngetRes(armletData->triSawTable);
	UngetRes(armletData->triRectTable);
	UngetRes(armletData->sawRectTable);
	UngetRes(armletData->triSawRectTable);
	UngetRes(armletData->egTable);
	UngetRes(armletData->egdrShift);
	UngetRes(armletData->sampleTab);
	UngetRes(armletData->sintab);
	UngetRes(armletData->resonanceLP);
	UngetRes(armletData->resonanceHP);

	for (i=0; i<12; i++)
		UngetRes(armletData->errors1541[i]);

	UncopyRes(armletData->prefs);

	MemPtrFree(armletData);
	armletData=NULL;
}
DumpArmlet()
{
	ArmletData *a=armletData;
	
	sprintf(msg, "Armlet %lx, ArmletData %lx.", armlet, a);
	SysFatalAlert(msg);
	sprintf(msg, "Ram1 %lx Ram2 %lx", a->ram1, a->ram2);
	SysFatalAlert(msg);
	sprintf(msg, "Kernal %lx Basic %lx char %lx color %lx",
		a->kernalROM, a->basicROM, a->charROM, a->color);
	SysFatalAlert(msg);
	sprintf(msg, "ram1541 %lx rom1541 %lx sintab %lx triTable %lx",
		a->ram1541, a->rom1541, a->sintab, a->triTable);
	SysFatalAlert(msg);
	sprintf(msg, "triSawTable %lx  %lx  %lx  %lx",
		a->triSawTable, a->triRectTable, a->sawRectTable, a->triSawRectTable);
	SysFatalAlert(msg);
	sprintf(msg, " %lx  %lx  %lx  %lx",
		a->egTable, a->egdrShift, a->sampleTab, a->resonanceLP);
	SysFatalAlert(msg);
	sprintf(msg, " %lx  %lx  %lx  %lx",
		a->resonanceHP, a->expTable, a->multiExpTable, a->textColorTable);
	SysFatalAlert(msg);
	sprintf(msg, " %lx  %lx  %lx",
		a->prefs, a->displayBuf, a->soundBuf);
	SysFatalAlert(msg);
	
	return 0;
}

// Paranoia City
CheckAlign(MemPtr p, char *name)
{
	if (((UInt32)p)&3) {
		sprintf(msg, "Pointer %s (0x%lx) isn't aligned!", p, name);
		SysFatalAlert(msg);
	}
	if (!p) {
		sprintf(msg, "Null pointer %s!", name);
		SysFatalAlert(msg);
	}
}
CheckAlignSwapped(MemPtr p, char *name)
{
	CheckAlign((MemPtr) swap32(p), name);
}
CheckArmlet()
{
	ArmletData *a=armletData;
	
	CheckAlign(armlet, "Armlet");
	CheckAlign(a, "ArmletData");

	CheckAlignSwapped(a->ram1, "RAM1");
	CheckAlignSwapped(a->ram2, "RAM1");
	CheckAlignSwapped(a->kernalROM, "Kernal ROM");
	CheckAlignSwapped(a->basicROM, "Basic ROM");
	CheckAlignSwapped(a->charROM, "Char ROM");
	CheckAlignSwapped(a->color, "Color");
	CheckAlignSwapped(a->ram1541, "RAM 1541");
	CheckAlignSwapped(a->rom1541, "ROM 1541");
	CheckAlignSwapped(a->triTable, "TriTable");
	CheckAlignSwapped(a->triSawTable, "TriSawTable");
	CheckAlignSwapped(a->triRectTable, "TriRectTable");
	CheckAlignSwapped(a->sawRectTable, "SawRectTable");
	CheckAlignSwapped(a->triSawRectTable, "TriSawRectTable");
	CheckAlignSwapped(a->egTable, "EGTable");
	CheckAlignSwapped(a->egdrShift, "EGDRShift");
	CheckAlignSwapped(a->sampleTab, "SampleTab");
	CheckAlignSwapped(a->sintab, "Sintab");
	CheckAlignSwapped(a->resonanceLP, "ResonanceLP");
	CheckAlignSwapped(a->resonanceHP, "ResonanceHP");
	CheckAlignSwapped(a->expTable, "ExpTable");
	CheckAlignSwapped(a->multiExpTable, "MultiExpTable");
	CheckAlignSwapped(a->textColorTable, "TextColorTable");
	CheckAlignSwapped(a->prefs, "Preferences");
	CheckAlignSwapped(a->displayBuf, "Display Buffer");
	CheckAlignSwapped(a->soundBuf, "Sound Buffer");
}

// Memory and resources 

// MemPtrNew with byteswapping
MemPtr NewSwapped(UInt32 n)
{
	MemPtr p=MemPtrNew(n);
	return (MemPtr) swap32(p);
}
void FreeSwapped(MemPtr p)
{
	MemPtr p2=(MemPtr) swap32(p);
	if (p2) MemPtrFree(p2);
}

// Load a resource and complain if it can't be found
MemPtr GetRes(DmResType t, UInt32 idx, char *name)
{
	MemHandle h;
	MemPtr p;

	h=DmGetResource(t, idx);
	if (!h) {
		sprintf(msg, "Couldn't find resource '%s' (%ld)", name, idx);
		SysFatalAlert(msg);
		return NULL;
	}
	p=MemHandleLock(h);
	if (!p) {
		DmReleaseResource(h);
		return NULL;
	}
	return (MemPtr) swap32(p);
}
void UngetRes(MemPtr p)
{
	MemHandle h;
	MemPtr p2;

	if (!p) return;

	p2=(MemPtr) swap32(p);

	h=MemPtrRecoverHandle(p2);
	if (!h) {
		sprintf(msg, "Couldn't recover handle for 0x%lx!", p2);
		SysFatalAlert(msg);
	}
	MemHandleUnlock(h);
	DmReleaseResource(h);
}

// Same as above, but copy the resource to dynamic (writable) memory
// (which also means that you can release the resource immideately)

// Loud version
MemPtr CopyRes(DmResType t, UInt32 idx, char *name)
{
	return CopyRes2(t, idx, name, 0);
}
// Variable volume
MemPtr CopyRes2(DmResType t, UInt32 idx, char *name, int quiet)
{
	MemHandle h;
	MemPtr src, dst;

	h=DmGetResource(t, idx);
	if (!h) {
		if (!quiet) {
			sprintf(msg, "Couldn't find resource '%s'", name);
			SysFatalAlert(msg);
		}
		return NULL;
	}
	// Allocate memory, copy resource
	dst = MemPtrNew(MemHandleSize(h));
	if (!dst) {
		DmReleaseResource(h);
		if (!quiet) {
			sprintf(msg, "Out of memory loading '%s'", name);
			SysFatalAlert(msg);
		}
		return NULL;
	}
	src = MemHandleLock(h);
	MemMove(dst, src, MemHandleSize(h));
	MemHandleUnlock(h);
	DmReleaseResource(h);

	return (MemPtr) swap32(dst);
}
void UncopyRes(void *p)
{
	MemPtr p2=(MemPtr) swap32(p);

	if (p2) MemPtrFree(p2);
}

// Fix up displayBuf and soundBuf
UInt32 InitGraphicsSound()
{
	UInt32 error;
	UInt16 i16;
	UInt32 w=320, h=320, d=4;
	Boolean col=true;

	if (can16 && is320) {
		// Try drawing directly to the screen
		if (error=WinScreenMode(winScreenModeSet, &w, &h, &d, &col)) {
			sprintf(msg, "Couldn't change screen mode (%ld)",
				error);
			SysFatalAlert(msg);
			return error;
		}
		screenBits=(UInt8 *) BmpGetBits(WinGetBitmap(WinGetDrawWindow()));
		if (!screenBits) {
			sprintf(msg, "Couldn't get screen bits");
			SysFatalAlert(msg);
			return 1;
		}
		armletData->displayBuf = (UInt8 *) swap32(screenBits);
		// There's an off chance that the device will report double
		// density but idoes in fact have a width other than 320 pixels.
		// <320 was tackled in CheckPrerequisites, >320 right here.
		if (screenW>320) {
			UInt16 d=screenW-320;
			UInt16 scroll=(d<<2)&0xfffc; // Center, keep it aligned

			armletData->displayModulo=swap32(d);
			armletData->displayBuf+=scroll;
		}
	} else {
		// Wrong size or no 16-color mode. Render to bitmap.
		bmp=BmpCreate(320, 272, 4, (ColorTableType *) &colorTable, &i16);
		if (!bmp) {
			sprintf("Error %d in BmpCreate()", i16);
			SysFatalAlert(msg);
			return 1;
		}
		bmpv3=BmpCreateBitmapV3(bmp, kDensityDouble, BmpGetBits(bmp), NULL);
		if (!bmpv3) {
			SysFatalAlert("Couldn't create high-res Bitmap");
			return 1;
		}
		bmpBits=BmpGetBits((BitmapType *)bmpv3);
		armletData->displayBuf = (UInt8 *) swap32(bmpBits);
		armletData->displayModulo = 0;
	}
	// Change palette to C=64 colors
	// Yes, even if we're not rendering to the screen.
	// More efficient that way (in non-direct color modes).
	if (error=WinPalette(winPaletteSet, 0, 16, colorTable.pal)) {
		sprintf(msg, "Couldn't change palette (0x%lx)", error);
		SysFatalAlert(msg);
		return error;
	}

	// Now the sound
	// Set the buffer even if we won't use it (just in case...)
	armletData->soundBuf = (UInt16 *) swap32(&(armletBuf[0]));
	// TODO debug sound
	//SysFatalAlert("1");
	canSound=0;
	if (!canSound) {
		return 0;
	}

	// ..Extended is 5.1? 5.2? Not 5.0 at any rate
	/*error = SndStreamCreateExtended(&sndStream, sndOutput, sndFormatPCM, 44100,
		sndInt16Little, sndMono, SoundCallback, armletData, 882, false);
	*/
	/*
	error = SndStreamCreateExtended(&sndStream, sndOutput, 0, 44100,
		sndInt16Little, sndMono, SoundCallback, armletData, 882, false);
	*/
	// SoundCallback will be called whenever the hw needs samples.
	// We have NO WAY of knowing how many samples will be demanded,
	// and must deliver them STAT. Oh joy.
	error = SndStreamCreate(&sndStream, sndOutput, 44100,
		sndInt16Little, sndMono, SoundCallback, &soundStruct,
			FRAME_SAMPLES, false);
	
	if (error) {
		SysFatalAlert("Couldn't start sound");
	} else {
		if (prefs->Sound) {
			SndStreamStart(sndStream);
		}
	}
	return 0;
}

// Shut of audio, release bitmaps, reset scrren mode
FreeGraphicsSound()
{
	if (!armlet) return;

	if (screenBits) {
		WinScreenMode(winScreenModeSetToDefaults, NULL,NULL,NULL,NULL);
	} else if (bmpBits) {
		BmpDelete(bmp);
		BmpDelete((BitmapType *) bmpv3);
	}
	if (sndStream) SndStreamDelete(sndStream);
}
/* Explanation: (See also VBlank) the SoundStruct contains a large(ish)
   ringbuffer of samples. SampleOffset is the index of the next sample that
   hasn't been output. SamplesQueued inidicates how many samples have been
   produced by the armlet beyond that point.
   SamplesQueued can never grow larger than the size of the buffer
   (if it does the data is thrown away rather than being queued.)
   The parameter 'size' is how many samples we must write to bufferP.
   Since I just fill any underflow with zeros, the net effect of wildly
   fluctuating frame rates should be that the latency of the stream 
   automatically goes up (amid horrible clicking).
   After a while the audio should be nice again (but high latency).
   This will ONLY work if the buffer size is larger than the maximum lag!
   Currently 4kB=0.1 s, kinda low.
   In case of pause, modal dialogs or the like the ringbuffer will quickly be
   emptied, and the whole process starts over when the emulator starts running
   normally again.
   If the Palm simply isn't fast enough to run at an average speed of 50HZ
   (even when dropping 90% of the frames) the sound WILL be shot to hell.
   That's what the 'Mute' option is for.
*/
Err SoundCallback(void *userDataP, SndStreamRef stream,
                  void *bufferP, UInt32 size)
{
	static int first=1, n=0;
	struct SoundStruct *ss=(struct SoundStruct *) userDataP;
	UInt16 *p=(UInt16 *)bufferP;
	UInt32 nsmpl=size, nempty=0;
	int i;
	char msg[101];

	/*
	MemSet(p, size, 0x80);
	MemSet(p+size/2, size, 0x7f);
	*/

	for (i=0; i<size; i++) {
		if (n++<100)
			p[i]=0x8000;
		else
			p[i]=0x7fff;
		if (n>200) n-=200;
	}

	return errNone;
	if (size > ss->samplesQueued) {
		// Underflow, fill with 0 
		nempty = size-ss->samplesQueued;
		nsmpl = ss->samplesQueued;
	}
	if (ss->sampleOffset+nsmpl > SOUND_BUFSIZE) {
		MemMove(p, ss->soundBuf+ss->sampleOffset,
			(SOUND_BUFSIZE - ss->sampleOffset)*2);
		MemMove(p, ss->soundBuf, 
			(ss->sampleOffset+nsmpl - SOUND_BUFSIZE)*2);
		ss->sampleOffset += nsmpl;
		ss->sampleOffset -= SOUND_BUFSIZE;
	} else {
		MemMove(p, ss->soundBuf+ss->sampleOffset, nsmpl*2);
		ss->sampleOffset += nsmpl;
	}
	MemSet(p+nsmpl, nempty*2, 0x0);
	ss->samplesQueued -= nsmpl;
	return errNone;
}

// Finally, go ARM!
UInt32 RunArmlet()
{
	UInt32 error;
	
	//SysFatalAlert("3");
	FrmGotoForm(MainForm);
	//SysFatalAlert("4");
	//DumpArmlet();
#ifdef DEBUG_GUI

	// Pretend armlet is running
	while (!armletData->errCode)
		EventLoop();
	/*
	while (!armletData->errCode) {
		VBlank68k(armletData);
	}
	*/
#else
	error=PceNativeCall((MemPtr)armlet, armletData);
#endif
	return armletData->errCode;
}

// Events

EventLoop()
{
	static UInt32 frame=0, cnt=0, tps, tpf, resets=0;
	EventType e;
	UInt32 startTick, tick, diff, sleepTo=0;
	UInt16 fillAt;
	static UInt32 lastTick;
	static UInt8  nskipped=0;
	Boolean handled;
	UInt16 menuError;

	// Get current time
	// I do this first to get maximum precision in the clocks.
	// sleepTo is when this frame should end, in a perfect world.
	// It's lastTick+1/50 s. If we lag to badly (10 frames right now),
	// the timing is reset.

	if (!lastTick || nskipped>=10) {
		// First time, or after pause/dialog, or just a very slow cpu.
		lastTick=tick=sleepTo=TimGetTicks();
		nskipped=0;
		resets++;
	} else {
		tick=TimGetTicks();
		sleepTo=lastTick+tpf;
	}
	if (frame==0L) {
		tps=SysTicksPerSecond();
		tpf=tps/50;
		startTick=tick;
	} /*else if (frame==500) {
		sprintf(msg, "%ld frames in %ld ticks (limit=%ld adaptive=%ld, "\
				"%ld tps -> %ld tpf-> %ld ms, %ld events, %ld resets)",
			frame, tick-startTick, prefs->LimitSpeed, prefs->AdaptiveFrameSkip,
			tps, tpf, (tick-startTick)*1000/tps, cnt, resets);
		SysFatalAlert(msg);
		startTick=tick=lastTick=sleepTo=TimGetTicks();
		prefs->LimitSpeed=0;
	} else if (frame==1000) {
		sprintf(msg, "%ld frames in %ld ticks (limit=%ld adaptive=%ld, "\
				"%ld tps -> %ld tpf-> %ld ms, %ld events, %ld resets)",
			frame-500, tick-startTick, prefs->LimitSpeed, prefs->AdaptiveFrameSkip,
			tps, tpf, (tick-startTick)*1000/tps, cnt, resets);
		SysFatalAlert(msg);
	}
	*/
	frame++;

	// Display
	if (armletData->drawFrame && !screenBits) {
		// Not rendering directly to screen, so blit bitmap
		// TODO: center this hor. on devices with weird widths
		// (scaling should be handled automagically)
		WinDrawBitmap(bmp, 0, 0);
	}
	// Prefs
	if (armletData->getPrefs) {
		// We ordered a set of prefs, well, here they are!
		armletData->getPrefs=0;
		SavePrefs();
	}

	// More timing
	// Send hint to armlet to skip next frame if we're running late
	
	armletData->drawFrame=0L;
	if (!prefs->LimitSpeed) {
		// Turbo mode, timing off
		sleepTo=tick;
	} else if (prefs->AdaptiveFrameSkip) {
		// Adaptive mode, skip next frame if we're running late.
		if (tick <= sleepTo)
			armletData->drawFrame=1;
	} else {
		// Old-style, use user-specified frame rates.
		if (nskipped+1 >= swap32(prefs->SkipFrames))
			armletData->drawFrame=1L;
	}
	if (armletData->drawFrame)
		nskipped=0;
	else
		nskipped++;
	
	// Sound, don't overqueue in turbo mode
	if (canSound && prefs->Sound &&
	    soundStruct.samplesQueued+FRAME_SAMPLES<SOUND_BUFSIZE) {
		// Explanation:
		//   soundBuf is a large(ish) circular buffer that's filled here
		//   with data from the armlet and drained by the sound callback.
		//   SampleOffset is the offset (in samples) to the next sample
		//   that will be played. SamplesQueued is the number of samples
		//   beyond this point that are ready for consumption.
		
		fillAt=soundStruct.sampleOffset+soundStruct.samplesQueued;
		if (fillAt+FRAME_SAMPLES > SOUND_BUFSIZE) {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				(SOUND_BUFSIZE-fillAt)*2);
			MemMove(soundStruct.soundBuf,
				armletBuf+SOUND_BUFSIZE-fillAt,
				(fillAt+FRAME_SAMPLES-SOUND_BUFSIZE)*2);
		} else {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				FRAME_SAMPLES*2);
		}
		soundStruct.samplesQueued+=FRAME_SAMPLES;
	}

	// Handle queued up events
	do {
		UInt16 menuError;
		
		tick=TimGetTicks();
		if (sleepTo-tick > SysTicksPerSecond()) {
			// Something very strange just happened....
			// int32 roll-over? Second coming of Christ?
			sleepTo=tick;
		}
		// Get event, sleepTo-tick is the max we should wait
		if (tick >= sleepTo)
			EvtGetEvent(&e, 0);
		else
			EvtGetEvent(&e, sleepTo-tick);
		if (e.eType==nilEvent) break;

		cnt++;
		handled=0;
		
		//EvtGetEvent(&e, 20);

		// Check pen position, I don't trust non-pen events.
		if (e.eType==penDownEvent) {
			penDown=true;
			penX=e.screenX;
			penY=e.screenY;
		}
		if (e.eType==penUpEvent)
			penDown=false;
		if (e.eType==penMoveEvent) {
			penX=e.screenX;
			penY=e.screenY;
		}

		if (e.eType==keyDownEvent) 
			HandleKey(&e);

		// Standard Palm Event handling
		if (!SysHandleEvent(&e))
			if (!MenuHandleEvent(0,&e,&menuError))
				if (!AppHandleEvent(&e)) // That's me
					FrmDispatchEvent(&e);
	} while (e.eType!=appStopEvent);
	
	if (e.eType==appStopEvent) armletData->errCode=ERR_NORMAL_EXIT;
	
	lastTick=sleepTo;
}

static void ShowDialog(UInt32 d)
{
	FormType *f=FrmInitForm(d);
	
	FrmDoDialog(f);
	FrmDeleteForm(f);
}
	
static Boolean HandleEventMain(EventPtr e)
{
	Boolean handled=false;
	UInt32 frmid;
	FormType *frm;

	switch (e->eType) {
		case menuEvent:
			handled=true;
			switch (e->data.menu.itemID) {
				case OptionsAboutFrodo:
					ShowDialog(AboutForm); 
					break;
				case OptionsJoystick:
					ShowDialog(JoystickForm);
					break;
				case OptionsObscure:
					ShowDialog(ObscureForm);
					break;
				default:
					handled=false;
			}
			break;
		case frmOpenEvent:
			frm=FrmGetActiveForm();
			FrmDrawForm(frm);
			handled=true;
			break;
	}
	return handled;
}
static Boolean HandleEventAbout(EventPtr e)
{
	Boolean handled=false;
	UInt32 frmid;
	FormType *frm;

	switch (e->eType) {
		case frmOpenEvent:
			frm=FrmGetActiveForm();
			FrmDrawForm(frm);
			handled=true;
			break;
	}
	return handled;
}
static Boolean HandleEventJoystick(EventPtr e)
{
	Boolean handled=false;
	UInt32 frmid;
	FormType *frm;

	switch (e->eType) {
		case frmOpenEvent:
			frm=FrmGetActiveForm();
			FrmDrawForm(frm);
			handled=true;
			break;
	}
	return handled;
}
static Boolean HandleEventObscure(EventPtr e)
{
	Boolean handled=false;
	UInt32 frmid;
	FormType *frm;

	switch (e->eType) {
		case frmOpenEvent:
			frm=FrmGetActiveForm();
			FrmDrawForm(frm);
			handled=true;
			break;
	}
	return handled;
}



static Boolean AppHandleEvent(EventPtr e)
{
	Boolean handled=false;
	FormType *frm;

	switch (e->eType) {
		case frmLoadEvent:
			frm=FrmInitForm(e->data.frmLoad.formID);
			FrmSetActiveForm(frm);
			switch (e->data.frmLoad.formID) {
				case MainForm:
					FrmSetEventHandler(frm, HandleEventMain);
					break;
				case AboutForm:
					FrmSetEventHandler(frm, HandleEventAbout);
					break;
				case JoystickForm:
					FrmSetEventHandler(frm,
						HandleEventJoystick);
					break;
				case ObscureForm:
					FrmSetEventHandler(frm,
						HandleEventObscure);
					break;
			}
			handled=true;
			break;
		case frmOpenEvent:
			frm=FrmGetActiveForm();
			FrmDrawForm(frm);
			handled=true;
			break;
		case appStopEvent:
			frm = FrmGetActiveForm();
			FrmEraseForm(frm);
			FrmDeleteForm(frm);
			handled = true;
			break;
	}
	return handled;
}

// ARM -> 68k callbacks

// Called every frame by the armlet. Our only chance to handle all kinds of I/O
UInt32 VBlank68k(void *p)
{
	static UInt32 frame=0, cnt=0;
	EventType e;
	UInt32 startTick, tick, diff, sleepTo=0;
	UInt16 fillAt;
	static UInt32 lastTick;
	static UInt8  nskipped=0;
	Boolean handled;

	//SysFatalAlert("5");
	
	// Get current time
	// I do this first to get maximum precision in the clocks.
	// sleepTo is when this frame should end, in a perfect world.
	// It's lastTick+1/50 s. If we lag to badly (10 frames right now),
	// the timing is reset.
/*
	if (!lastTick || nskipped>=10) {
		// First time, or after pause/dialog, or just a very slow cpu.
		lastTick=tick=sleepTo=TimGetTicks();
	} else {
		tick=TimGetTicks();
		sleepTo=lastTick+SysTicksPerSecond()/50;
	}
	if (frame==0) {
		startTick=tick;
	} else if (frame==100 || cnt==200) {
		sprintf(msg, "%d frames in %ld ticks (%ld events)",
			frame, tick-startTick, cnt);
		SysFatalAlert(msg);
	}
	frame++;
*/
/*
	// Display
	if (armletData->drawFrame && !screenBits) {
		// Not rendering directly to screen, so blit bitmap
		// TODO: center this hor. on devices with weird widths
		// (scaling should be handled automagically)
		WinDrawBitmap(bmp, 0, 0);
	}
	// Prefs
	if (armletData->getPrefs) {
		// We ordered a set of prefs, well, here they are!
		armletData->getPrefs=0;
		SavePrefs();
	}

	// More timing
	// Send hint to armlet to skip next frame if we're running late
	if (!prefs->LimitSpeed) {
		// Turbo mode, timing off
		sleepTo=tick;
		armletData->drawFrame=1;
	} else if (prefs->AdaptiveFrameSkip) {
		// Adaptive mode, skip next frame if we're running late.
		armletData->drawFrame = tick<=sleepTo;
	} else {
		// Old-style, use user-specified frame rates.
		armletData->drawFrame=0;
		if (++nskipped >= swap32(prefs->SkipFrames)) {
			nskipped=0;
			armletData->drawFrame=1;
		}
	}
*/
/*

	// Sound, don't overqueue in turbo mode
	if (canSound && prefs->Sound &&
	    soundStruct.samplesQueued+FRAME_SAMPLES<SOUND_BUFSIZE) {
		// Explanation:
		//   soundBuf is a large(ish) circular buffer that's filled here
		//   with data from the armlet and drained by the sound callback.
		//   SampleOffset is the offset (in samples) to the next sample
		//   that will be played. SamplesQueued is the number of samples
		//   beyond this point that are ready for consumption.
		
		fillAt=soundStruct.sampleOffset+soundStruct.samplesQueued;
		if (fillAt+FRAME_SAMPLES > SOUND_BUFSIZE) {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				(SOUND_BUFSIZE-fillAt)*2);
			MemMove(soundStruct.soundBuf,
				armletBuf+SOUND_BUFSIZE-fillAt,
				(fillAt+FRAME_SAMPLES-SOUND_BUFSIZE)*2);
		} else {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				FRAME_SAMPLES*2);
		}
		soundStruct.samplesQueued+=FRAME_SAMPLES;
	}
*/

	// Handle queued up events
	do {
		UInt16 menuError;
		tick=TimGetTicks();
		if (sleepTo-tick > SysTicksPerSecond()) {
			// Something very strange just happened....
			// int32 roll-over? Second coming of Christ?
			sleepTo=tick;
		}
		// Get event, sleepTo-tick is the max we should wait
		if (tick >= sleepTo)
			EvtGetEvent(&e, 0);
		else
			EvtGetEvent(&e, sleepTo-tick);
		if (e.eType==nilEvent) break;

		cnt++;
		handled=0;

		// Check pen position, I don't trust non-pen events.
		if (e.eType==penDownEvent) {
			penDown=true;
			penX=e.screenX;
			penY=e.screenY;
		}
		if (e.eType==penUpEvent)
			penDown=false;
		if (e.eType==penMoveEvent) {
			penX=e.screenX;
			penY=e.screenY;
		}

		if (e.eType=keyDownEvent) 
			HandleKey(&e);

		// Standard Palm Event handling
		if (!SysHandleEvent(&e))
			if (!MenuHandleEvent(0,&e,&menuError))
				if (!AppHandleEvent(&e)) // That's me
					FrmDispatchEvent(&e);

	} while (e.eType!=appStopEvent && e.eType!=nilEvent);

	// Joystick
	armletData->Joystick=0;
/*
	CheckJoyPen();
	CheckJoyKey();
*/

	// TODO: Keyboard

	if (e.eType==appStopEvent)
		armletData->errCode=ERR_NORMAL_EXIT;
	
	lastTick=sleepTo;
	return 0;
}

// The JoyPen joystick emulation uses the pen to specify jostick direction.
// Basically, you have an origin and and extent, then see which octant the
// pen is in.

UInt32 CheckJoyPen2(uint32 x, uint32 y, uint32 sz)
{
	Int32 dx=penX-x-(sz>>1), dy=penY-y-(sz>>1);
	Int32 ax=dx, ay=dy;
	UInt32 center=sz>>4;

	// In the zone?
	if (!penDown || penX<x || penY<y || penX>=x+sz || penY>=y+sz)
		return;

	// ax,ay = abs(dx), abs(dy)
	if (ax<0) ax=-ax;
	if (ay<0) ay=-ay;

	// If we're in the inner 16'th then call it centered joystick.
	if (center==0) center=1;
	if (ax<center && ay<center) return 0;

	// tan v=sin v/cos v = y/x
	// tan pi/8 = y/x = 0.4142
	// x*4142 = y*10000
	// Ie. in the below means "In the left-right octants"
	if (ax*4142 > ay*10000) {
		// Left, right
		if (dx<0)
			return swap32(JOYLEFT);
		else
			return swap32(JOYRIGHT);
	} else if (ay*4142 > ax*10000) {
		// Up, down
		if (dy<0)
			return swap32(JOYUP);
		else
			return swap32(JOYDOWN);
	} else if (dx<0) {
		// UpLeft, DownLeft
		if (dy<0)
			return swap32(JOYUP | JOYLEFT);
		else
			return swap32(JOYDOWN | JOYLEFT);
	} else {
		// UpRight, DownRight
		if (dy<0)
			return swap32(JOYUP | JOYRIGHT);
		else
			return swap32(JOYDOWN | JOYRIGHT);
	}
}
//
CheckJoyPen()
{
	if (!prefs->JoyPen) return;

	if (prefs->JoyPen & swap32(JOYPEN_SCREEN)) {
		armletData->Joystick |= CheckJoyPen2(0,0,160);
	} else if (prefs->JoyPen & swap32(JOYPEN_GRAFFITI)) {
		// TODO: the condensed joypen mode
	}
}

// Joystick keys on the keyboard
CheckJoyKey()
{
	UInt32 u, l, r, d, f, k, j;

	// Directions may have been partially set by JoyPen
	j=swap32(armletData->Joystick);
	u=swap32(prefs->KeyUp);
	l=swap32(prefs->KeyLeft);
	r=swap32(prefs->KeyRight);
	d=swap32(prefs->KeyDown);
	f=swap32(prefs->KeyFire);

	k=KeyCurrentState();

	if (k&u) j|=JOYUP;
	if (k&l) j|=JOYLEFT;
	if (k&r) j|=JOYRIGHT;
	if (k&d) j|=JOYDOWN;
	if (k&f) j|=JOYFIRE;

	armletData->Joystick=swap32(j);
}

// forward graffiti chars to the emulator
HandleKey(EventType *e)
{
	// TODO: implement graffiti input
}

SavePrefs()
{
	MemHandle h;
	MemPtr p;
	int i;

	if (!prefs) {
		SysFatalAlert("No prefs to save...");
		return 5;
	}
	// Check if data is all 0. If it is, somehow a nuked prefs made it here.
	if (IsNukedPrefs(prefs)) {
		SysFatalAlert("Prefs is nuked, won't save");
		return 6;
	}

	if (!(h=DmGetResource(prefsType, IDX_PREFS))) {
		SysFatalAlert("Error saving preferences");
		return 1;
	}
	if (MemHandleSize(h) < sizeof(struct PrefsState))  {
		SysFatalAlert("Prefs resource to small?!");
		return 2;
	} 
	if (!(p=MemHandleLock(h))) {
		SysFatalAlert("Couldn't lock prefs!");
		return 3;
	} 
	// And finally write the prefs.
	if (DmWrite(p, 0, prefs, sizeof(struct PrefsState))
			!= errNone) {
		MemHandleUnlock(h);
		SysFatalAlert("Error writing preferences");
		return 4;
	}

	MemHandleUnlock(h);
	return 0;
}

IsNukedPrefs(struct PrefsState *p)
{
	UInt8 *p2=(UInt8 *) p;
	UInt32 i;
	
	for (i=0; i<sizeof(struct PrefsState); i++) 
		if (p2[i]!=0x0) break;
		
	return i<sizeof(struct PrefsState);
}

// Display one of the requesters. Just ugly FatalAlerts right now.
UInt32 ShowRequester68k(UInt32 *p)
{
	UInt32 id=swap32(p[0]);
	UInt32 arg1=swap32(p[1]), arg2=swap32(p[2]);
	UInt32 arg3=swap32(p[3]), arg4=swap32(p[4]);

	switch (id) {
		case REQ_BAD_SNAPSHOT:
			sprintf(msg, "Not a valid Frodo snapshot.");
			break;
		case REQ_BAD_SNAPSHOT_VERSION:
			sprintf(msg, "Invalid snapshot version.");
			break;
		case REQ_BAD_OPCODE:
			sprintf(msg, "Illegal opcode 0x%04x at 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP:
			sprintf(msg, "Jump to I/O space at 0x%04x to 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_OPCODE_1541:
			sprintf(msg, "1541: Illegal opcode 0x%04x at 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP_1541:
			sprintf(msg, "1541: Jump to I/O space at 0x%04x to 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_NEW_DRIVE:
			sprintf(msg, "Out of memory changing disk %d.",
				((UInt16)arg1));
			break;
		case REQ_NEW_SIDRENDERER:
			sprintf(msg, "Out of memory when starting audio.");
			break;
		deafult:
			sprintf(msg, "Unknow req %ld (%ld,%ld,%ld,%ld)",
				arg1, arg2, arg3, arg4);
	}
	// TODO 1541 errors should have two buttons "Reset C64", "Reset 1541"
	// I really should draw real Alrt's for these...
	SysFatalAlert(msg);

	return 0;
}

void *Malloc68k(UInt32 n)
{
	return MemPtrNew(n);
}

UInt32 Free68k(void *p)
{
	MemPtrFree(p);
	return 0;
}

/*
static UWORD ComputeX( UWORD value )
{
	return(( UWORD )((( FontX * value ) + 4 ) / 8 ));
}

static UWORD ComputeY( UWORD value )
{
	return(( UWORD )((( FontY * value ) + 4 ) / 8 ));
}

static void ComputeFont( UWORD width, UWORD height )
{
	Font = &Attr;
	Font->ta_Name = (STRPTR)Scr->RastPort.Font->tf_Message.mn_Node.ln_Name;
	Font->ta_YSize = FontY = Scr->RastPort.Font->tf_YSize;
	FontX = Scr->RastPort.Font->tf_XSize;

	OffX = Scr->WBorLeft;
	OffY = Scr->RastPort.TxHeight + Scr->WBorTop + 1;

	if ( width && height ) {
		if (( ComputeX( width ) + OffX + Scr->WBorRight ) > Scr->Width )
			goto UseTopaz;
		if (( ComputeY( height ) + OffY + Scr->WBorBottom ) > Scr->Height )
			goto UseTopaz;
	}
	return;

UseTopaz:
	Font->ta_Name = (STRPTR)"topaz.font";
	FontX = FontY = Font->ta_YSize = 8;
}

int SetupScreen( void )
{
	if ( ! ( Scr = LockPubScreen( PubScreenName )))
		return( 1L );

	ComputeFont( 0, 0 );

	if ( ! ( VisualInfo = GetVisualInfo( Scr, TAG_DONE )))
		return( 2L );

	return( 0L );
}

void CloseDownScreen( void )
{
	if ( VisualInfo ) {
		FreeVisualInfo( VisualInfo );
		VisualInfo = NULL;
	}

	if ( Scr        ) {
		UnlockPubScreen( NULL, Scr );
		Scr = NULL;
	}
}

void PrefsRender( void )
{
	struct IntuiText	it;
	UWORD			cnt;

	ComputeFont( PrefsWidth, PrefsHeight );


	for ( cnt = 0; cnt < Prefs_TNUM; cnt++ ) {
		CopyMem(( char * )&PrefsIText[ cnt ], ( char * )&it, (long)sizeof( struct IntuiText ));
		it.ITextFont = Font;
		it.LeftEdge  = OffX + ComputeX( it.LeftEdge ) - ( IntuiTextLength( &it ) >> 1 );
		it.TopEdge   = OffY + ComputeY( it.TopEdge ) - ( Font->ta_YSize >> 1 );
		PrintIText( PrefsWnd->RPort, &it, 0, 0 );
	}
}

int HandlePrefsIDCMP( void )
{
	struct IntuiMessage	*m;
	struct MenuItem		*n;
	int			(*func)();
	BOOL			running = true;

	while( m = GT_GetIMsg( PrefsWnd->UserPort )) {

		CopyMem(( char * )m, ( char * )&PrefsMsg, (long)sizeof( struct IntuiMessage ));

		GT_ReplyIMsg( m );

		switch ( PrefsMsg.Class ) {

			case	IDCMP_REFRESHWINDOW:
				GT_BeginRefresh( PrefsWnd );
				PrefsRender();
				GT_EndRefresh( PrefsWnd, TRUE );
				break;

			case	IDCMP_VANILLAKEY:
				running = PrefsVanillaKey();
				break;

			case	IDCMP_GADGETUP:
				func = ( void * )(( struct Gadget * )PrefsMsg.IAddress )->UserData;
				running = func();
				break;

			case	IDCMP_MENUPICK:
				while( PrefsMsg.Code != MENUNULL ) {
					n = ItemAddress( PrefsMenus, PrefsMsg.Code );
					func = (void *)(GTMENUITEM_USERDATA( n ));
					running = func();
					PrefsMsg.Code = n->NextSelect;
				}
				break;
		}
	}
	return( running );
}

int OpenPrefsWindow( void )
{
	struct NewGadget	ng;
	struct Gadget	*g;
	UWORD		lc, tc;
	UWORD		wleft = PrefsLeft, wtop = PrefsTop, ww, wh;

	ComputeFont( PrefsWidth, PrefsHeight );

	ww = ComputeX( PrefsWidth );
	wh = ComputeY( PrefsHeight );

	if (( wleft + ww + OffX + Scr->WBorRight ) > Scr->Width ) wleft = Scr->Width - ww;
	if (( wtop + wh + OffY + Scr->WBorBottom ) > Scr->Height ) wtop = Scr->Height - wh;

	if ( ! ( g = CreateContext( &PrefsGList )))
		return( 1L );

	for( lc = 0, tc = 0; lc < Prefs_CNT; lc++ ) {

		CopyMem((char * )&PrefsNGad[ lc ], (char * )&ng, (long)sizeof( struct NewGadget ));

		ng.ng_VisualInfo = VisualInfo;
		ng.ng_TextAttr   = Font;
		ng.ng_LeftEdge   = OffX + ComputeX( ng.ng_LeftEdge );
		ng.ng_TopEdge    = OffY + ComputeY( ng.ng_TopEdge );
		ng.ng_Width      = ComputeX( ng.ng_Width );
		ng.ng_Height     = ComputeY( ng.ng_Height);

		PrefsGadgets[ lc ] = g = CreateGadgetA((ULONG)PrefsGTypes[ lc ], g, &ng, ( struct TagItem * )&PrefsGTags[ tc ] );

		while( PrefsGTags[ tc ] ) tc += 2;
		tc++;

		if ( NOT g )
			return( 2L );
	}

	if ( ! ( PrefsMenus = CreateMenus( PrefsNewMenu, GTMN_FrontPen, 0L, TAG_DONE )))
		return( 3L );

	LayoutMenus( PrefsMenus, VisualInfo, TAG_DONE );

		PrefsZoom[0] = PrefsZoom[1] = 0;
	if ( PrefsWdt )
		PrefsZoom[2] = TextLength( &Scr->RastPort, (UBYTE *)PrefsWdt, strlen((char *)PrefsWdt )) + 80;
	else
		PrefsZoom[2]  = 80L;
		PrefsZoom[3] = Scr->WBorTop + Scr->RastPort.TxHeight + 1;

	if ( ! ( PrefsWnd = OpenWindowTags( NULL,
				WA_Left,	wleft,
				WA_Top,		wtop,
				WA_Width,	ww + OffX + Scr->WBorRight,
				WA_Height,	wh + OffY + Scr->WBorBottom,
				WA_IDCMP,	CHECKBOXIDCMP|INTEGERIDCMP|CYCLEIDCMP|STRINGIDCMP|BUTTONIDCMP|IDCMP_MENUPICK|IDCMP_VANILLAKEY|IDCMP_REFRESHWINDOW,
				WA_Flags,	WFLG_DRAGBAR|WFLG_DEPTHGADGET|WFLG_SMART_REFRESH|WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE,
				WA_Gadgets,	PrefsGList,
				WA_Title,	PrefsWdt,
				WA_ScreenTitle,	"Frodo C64 Emulator",
				WA_PubScreen,	Scr,
				WA_Zoom,	PrefsZoom,
				TAG_DONE )))
	return( 4L );

	SetMenuStrip( PrefsWnd, PrefsMenus );
	GT_RefreshWindow( PrefsWnd, NULL );

	PrefsRender();

	return( 0L );
}

void ClosePrefsWindow( void )
{
	if ( PrefsMenus      ) {
		ClearMenuStrip( PrefsWnd );
		FreeMenus( PrefsMenus );
		PrefsMenus = NULL;	}

	if ( PrefsWnd        ) {
		CloseWindow( PrefsWnd );
		PrefsWnd = NULL;
	}

	if ( PrefsGList      ) {
		FreeGadgets( PrefsGList );
		PrefsGList = NULL;
	}
}
*/
