/*
 * PalmOSGUI.c
 * The GUI for the Palm. This source is 68k, unlike all the other files
 * which are compiled for ARM.
 */

#include <PalmOS.h>
#include <Bitmap.h>
#include <MemoryMgr.h>
#include <DataMgr.h>
#include <SoundMgr.h>
#include <PceNativeCall.h>
#include <Rect.h>

#include "PalmOSGUI.h"
#include "Armlet68k.h"
#include "FrodoResources.h"

#define IsNuked(a) (!StrNCompare((char *) (a), "NUKE", 4)) 
#define DEFAULT_SNAP_NAME "Frodo default snapshot"

static UInt16 ShowDialog(UInt16);

//#define DEBUG_GUI

UInt8 *armlet=NULL;
ArmletData *armletData=NULL;

#define SOUND_BUFSIZE 22050L
#define FRAME_SAMPLES 882

// Sound
UInt16 armletBuf[FRAME_SAMPLES]; // 1 frame of samples
struct SoundStruct {
	UInt32 sampleOffset, samplesQueued;
	Int32 underflow;
	UInt16 *soundBuf; //[SOUND_BUFSIZE]; // Ringbuffer for sound
	Boolean makeNoise;
	UInt32 noiseOffs;
} soundStruct={0,0,0,NULL,false,0};
SndStreamRef sndStream=NULL;
// Capabilities
UInt8 canSound=0, can16=0, can256=0, is320=0;
UInt32 screenW=0, screenH=0;
// Various
Boolean crashed=false;
Boolean firstTime=false;
Int32 penX, penY;
Boolean penDown=false;
// Our preferences (ARM byte order)
struct PrefsState *prefs;
// The disk/snapshot database
LocalID mainDB;
DmOpenRef mainDBRef;
// The (max) 4 mounted disks (8,9,10,11)
FILE disk[4];
UInt8 activeDrive=0; // The drive unit (C64 unit - 8) that's active, ie. target for mounts
Boolean exiting=false, pause=false;
MemHandle joyHdl[9]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
BitmapPtr joyPtr[9]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
// Graffiti keys
UInt32 ticksLastKeyDown=0, savedGraffitiKey;
Boolean mixedCase=false;

char *frodoDir="/Palm/programs/Frodo";

// C=64 colors, see Prefs.*
struct {
	UInt16 numEntries;
	RGBColorType pal[16];
} colorTable =
{ 16,
{
	{ 0,   0,   0,   0},
	{ 1, 255, 255, 255},
	{ 2, 116,  67,  53},
	{ 3, 124, 172, 186},
	{ 4, 123,  72, 144},
	{ 5, 100, 151,  79},
	{ 6,  64,  50, 133},
	{ 7, 191, 205, 122},
	{ 8, 123,  91,  47},
	{ 9,  79,  69,   0},
	{10, 163, 114, 101},
	{11,  80,  80,  80},
	{12, 120, 120, 120},
	{13, 164, 215, 142},
	{14, 120, 106, 189},
	{15, 159, 159, 159}
}
};

// Be2le macros
#define swap32(x) (((((UInt32)(x))&0x000000ffL)<<24) | \
                   ((((UInt32)(x))&0x0000ff00L)<< 8) | \
                   ((((UInt32)(x))&0x00ff0000L)>> 8) | \
                   ((((UInt32)(x))&0xff000000L)>>24))
#define swap16(x) (((((UInt16)(x))&0x00ff)<<8) | \
                   ((((UInt16)(x))&0xff00)>>8))

// Graphics
UInt8 *screenBits=NULL, *bmpBits=NULL;
BitmapType *bmp=NULL;
BitmapTypeV3 *bmpv3=NULL;

// Scratch string
char msg[256];

// Startup routine
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt32 error=0;

	switch (cmd) {
		case sysAppLaunchCmdNormalLaunch:
			CheckCrash(); // Check, then set "crashed" flag in prc ('CRSH' 0)
			error = Launch();
			NoCrash(); // Clear "crashed" flag
			break;
		case sysAppLaunchCmdFind:
			//error = Search((FindParamsPtr) cmdPBP);
			break;
	}
	if (error) return error;
	return 0;
}

// Check if Frodo crashed last time it was run.
CheckCrash()
{
	MemHandle h=DmGetResource(crashType, IDX_CRASH);
	UInt8 *p;
	
	if (!h) {
		SysFatalAlert("Crash report info not found!!");
		crashed=true;
		return;
	}
	p=(UInt8 *) MemHandleLock(h);
	if (!p) {
		SysFatalAlert("Couldn't lock crash report info?!");
		crashed=true;
		return;
	}
	crashed=*p;
	DmWrite(p, 0, "\001", 1);
	MemHandleUnlock(h);
	DmReleaseResource(h);
}
// We made it out alive.
NoCrash()
{
	MemHandle h=DmGetResource(crashType, IDX_CRASH);
	UInt8 *p;
	
	if (!h) {
		SysFatalAlert("Crash report info not found (on exit)!!");
		crashed=true;
		return;
	}
	p=(UInt8 *) MemHandleLock(h);
	if (!p) {
		SysFatalAlert("Couldn't lock crash report info (on exit)?!");
		crashed=true;
		return;
	}
	DmWrite(p, 0, "\0", 1);
	MemHandleUnlock(h);
	DmReleaseResource(h);
}

UInt32 Launch()
{
	UInt32 error=0;
	
	// Structured programming 8-/
	if (!(error=CheckPrerequisites())) {
	  if (!(error=CreateArmlet())) {
	    if (!(error=CreateArmletData())) {
	      if (!(error=InitDB())) {
		if (!(error=InitGraphicsSound())) {
		  //SysFatalAlert("2");

		  error = RunArmlet();
		  if (armletData->cmd & CMD68K_SAVE)
		    // With it's dying breath, the armlet gave us a snapshot
		    SaveSnapshot(DEFAULT_SNAP_NAME, 0);
		  
		  switch (swap32(error)) {
			case ERR_NORMAL_EXIT:
				error=0; break;
			case ERR_NEW_FRODO:
				SysFatalAlert("Out of memory creating Frodo"); break;
			case ERR_NEW_C64:
				SysFatalAlert("Out of memory creating C64"); break;
			case ERR_NEW_CPU:
				SysFatalAlert("Out of memory creating CPU"); break;
			case ERR_NEW_JOB1541:
				SysFatalAlert("Out of memory creating 1541Job"); break;	
			case ERR_NEW_CPU1541:
				SysFatalAlert("Out of memory creating 1541CPU"); break;
			case ERR_NEW_VIC:
				SysFatalAlert("Out of memory creating VIC"); break;
			case ERR_NEW_SID:
				SysFatalAlert("Out of memory creating SID"); break;
			case ERR_NEW_CIA1:
				SysFatalAlert("Out of memory creating CIA1"); break;
			case ERR_NEW_CIA2:
				SysFatalAlert("Out of memory creating CIA2"); break;
			case ERR_NEW_IEC:
				SysFatalAlert("Out of memory creating IEC"); break;
			case ERR_NEW_REU:
				SysFatalAlert("Out of memory creating REU"); break;
			default:
				StrPrintF(msg, "Unknow error %ld", error);
				SysFatalAlert(msg);
		  }
		  FreeGraphicsSound();
		} //GfxSound
		FreeDB();
	      } // DB
	      FreeArmletData();
	    } else { // ArmletData
	      StrPrintF(msg, "Error in CreateArmletData (0x%lx)", error);
	      SysFatalAlert(msg);
	    }
	    FtrPtrFree(appSignature, 0); // Remove the armlet feature
	  } // Armlet
	}  // Prerequisites
	return error;
}

UInt32 CheckPrerequisites()
{
	UInt32 w,h,d, error=0;
	Boolean c;

	// OS5+?
	error = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &d);
	if (error || d < sysMakeROMVersion(5,0,0,0,0)) {
		StrPrintF(msg, "You need PalmOS 5.0 or greater! (You have 0x%lx)",
			d);
		SysFatalAlert(msg);
		return 2;
	}
	// ARM processor?
	error = FtrGet(sysFileCSystem, sysFtrNumProcessorID, &d);
	if (error || !sysFtrNumProcessorIsARM(d)) {
		StrPrintF(msg, "You need an ARM processor to run Frodo (Your processor is 0x%lx)", d);
		SysFatalAlert(msg);
		return 7;
	}
	// Hires support?
	error = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &d);
	if (error || d < 4) {
		StrPrintF(msg, "You need high-res support, "\
			"Window Manager 4+. (You have 0x%lx)", d);
		SysFatalAlert(msg);
		return 3;
	}
	// Hires?
	WinScreenGetAttribute(winScreenDensity, &d);
	if (d < kDensityDouble) {
		StrPrintF(msg, "Your device doesn't have a screen with "\
			"320x320+ density. (Your density is 0x%lx)", d);
		SysFatalAlert(msg);
		return 4;
	}
	if (d == kDensityDouble) is320 = 1;

	// Wider than 320?
	error=WinScreenGetAttribute(winScreenWidth, &screenW);
	if (error) {
		SysFatalAlert("Your Palm refuses to divulge its screen width!");
		return 9;
	}
	if (screenW<320) {
		StrPrintF(msg, "Your Palm said it was hires, but in fact it only "\
			"has a screen width of %ld (need 320).", screenW);
		SysFatalAlert(msg);
		return 10;
	}
	// Higher than 272?
	error=WinScreenGetAttribute(winScreenHeight, &screenH);
	if (error) {
		SysFatalAlert("Your Palm refuses to report its screen height!");
		return 11;
	}
	// Actually, right now I'm demanding 320 because I'm lazy and can't
	// bother to turn of the on-screen keyboard if there isn't room for it
	if (screenH<320) {
		StrPrintF(msg, "The screen size %ldx%ld isn't tall enough. "\
			"(Frodo needs a minimum of 320x320 screen area)", screenW, screenH);
		SysFatalAlert(msg);
		return 12;
	}

	// Some optional features
	// Native 16 color mode?
	c = true;
	error = WinScreenMode(winScreenModeGetSupportedDepths, &w, &h, &d, &c);
	if (!error && (d & 8))  // if x cols are supported, d&(x/2) != 0
		can16 = true;
	if (!error && (d & 128))
		can256=true;

	// Sampled sound?

	// Does NOT WORK AT ALL on 5.0 Tungsten T v1 with pre-production OS 5.0
	// (ie. my Palm device!)

	if (!(error = FtrGet(sysFileCSoundMgr, sndFtrIDVersion, &d)))
		canSound = true;
	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &d);
	if(d == 'Frg1') // Tungsten T. 
		canSound = true;
		
	return 0;
}

#define MAXCHUNKS 5
UInt32 CreateArmlet()
{
	MemHandle armletPart[MAXCHUNKS];
	MemPtr p;
	UInt32 armletSize, offs;
	UInt8 nParts, i;
	UInt32 error;

	// Reassemble armlet from resources and stuff it in feature memory
	// Armlet is stored as Arml IDX_ARMLET, +1, +2, +3, ...
	armletSize=0;
	for (i=0; i<MAXCHUNKS; i++) {
		armletPart[i] = DmGetResource(armletType, IDX_ARMLET+i);
		if (!armletPart[i]) break;
		armletSize+=MemHandleSize(armletPart[i]);
	}
	if (i>=MAXCHUNKS) {
		StrPrintF(msg, "Too many chunks in armlet (%d)", i);
		SysFatalAlert(msg);
		for (i=0; i<MAXCHUNKS; i++) DmReleaseResource(armletPart[i]);
		return 1;
	}
	if (i==0) return DmGetLastErr();
	nParts=i;
	
	//StrPrintF(msg, "Armlet size %ld in %d parts", armletSize, nParts);
	//SysFatalAlert(msg);
	// Free feature 0 just in case we exited without freeing it last time
	FtrPtrFree(appSignature, 0);
	if (error = FtrPtrNew(appSignature, 0, armletSize, (void **) &armlet)) {
		SysFatalAlert("Couldn't create feature");
		for (i=0; i<nParts; i++) DmReleaseResource(armletPart[i]);
		return error;
	}

	offs=0;
	// Write the parts to feature memory
	for (i=0; i<nParts; i++) {
		p = MemHandleLock(armletPart[i]);
		error=DmWrite(armlet, offs, p, MemHandleSize(armletPart[i]));
		MemHandleUnlock(armletPart[i]);
		if (error) {
			StrPrintF(msg, "Couldn't write feature (part %d)", i);
			SysFatalAlert(msg);
			for (i=0; i<nParts; i++)
				DmReleaseResource(armletPart[i]);
			return error;
		}
		offs += MemHandleSize(armletPart[i]);
	}
	for (i=0; i<nParts; i++) DmReleaseResource(armletPart[i]);
	return 0;
}
	
// Allocate ArmletData and fill it in (mind the endianness!)
UInt32 CreateArmletData()
{
	UInt32 error, x;
	UInt8 i;
	DmOpenRef db;
	MemHandle h;

	armletData = (ArmletData *) MemPtrNew(sizeof(ArmletData));
	if (!armletData) return memErrNotEnoughSpace;
	MemSet(armletData, sizeof(ArmletData), 0x0);
	
	// Callbacks, needed some workarounds to get the pointers in there.
	x = (UInt32) VBlank68k;
	armletData->vBlank    = (void *) swap32(x);
	x = (UInt32) ShowRequester68k;
	armletData->requester = (void *) swap32(x);
	x = (UInt32) Malloc68k;
	armletData->malloc    = (void *) swap32(x);
	x = (UInt32) Free68k;
	armletData->free      = (void *) swap32(x);
										
	// RAM & ROM
	armletData->ram1 = (UInt8 *) NewSwapped(0x8000); // 32 kB
	if (!armletData->ram1) return memErrNotEnoughSpace;
	armletData->ram2 = (UInt8 *) NewSwapped(0x8000); // 32 kB
	if (!armletData->ram2) return memErrNotEnoughSpace;
	armletData->kernalROM = CopyRes(romType, IDX_KERNAL, "KernalROM");
	if (!armletData->kernalROM) return memErrNotEnoughSpace;
	armletData->basicROM = GetRes(romType, IDX_BASIC, "BasicROM");
	if (!armletData->basicROM) return memErrNotEnoughSpace;
	armletData->charROM = GetRes(romType, IDX_CHAR, "CharROM");
	if (!armletData->charROM) return memErrNotEnoughSpace;
	armletData->color = (UInt8 *) NewSwapped(0x0400); // Color RAM
	if (!armletData->color) return memErrNotEnoughSpace;
	armletData->ram1541 = (UInt8 *) NewSwapped(0x0800); // Diskdrive RAM
	if (!armletData->ram1541) return memErrNotEnoughSpace;
	armletData->rom1541 = CopyRes(romType, IDX_1541, "1541ROM");
	if (!armletData->rom1541) return memErrNotEnoughSpace;

	// Various tables and precalculations
	armletData->expTable = GetRes(tblType, IDX_EXPTABLE, "ExpTable");
	if (!armletData->expTable) return memErrNotEnoughSpace;
	armletData->multiExpTable=GetRes(tblType, IDX_MULTIEXPTABLE,
		"MultiExpTable");
	if (!armletData->multiExpTable) return memErrNotEnoughSpace;
	armletData->textColorTable=GetRes(tblType, IDX_TEXTCOLORTABLE,
		"Text color lookup table");
	if (!armletData->textColorTable) return memErrNotEnoughSpace;
	armletData->triTable=GetRes(tblType, IDX_TRITABLE, "TriTable");
	if (!armletData->triTable) return memErrNotEnoughSpace;
	armletData->triSawTable=GetRes(tblType, IDX_TRISAWTABLE, "TriSawTable");
	if (!armletData->triSawTable) return memErrNotEnoughSpace;
	armletData->triRectTable=GetRes(tblType, IDX_TRIRECTTABLE,
		"TriRectTable");
	if (!armletData->triRectTable) return memErrNotEnoughSpace;
	armletData->sawRectTable=GetRes(tblType, IDX_SAWRECTTABLE,
		"SawRectTable");
	if (!armletData->sawRectTable) return memErrNotEnoughSpace;
	armletData->triSawRectTable=GetRes(tblType, IDX_TRISAWRECTTABLE,
		"TriSawRectTable");
	if (!armletData->triSawRectTable) return memErrNotEnoughSpace;
	armletData->egTable=GetRes(tblType, IDX_EGTABLE, "EGTable");
	if (!armletData->egTable) return memErrNotEnoughSpace;
	armletData->egdrShift=GetRes(tblType, IDX_EGDRSHIFT, "EGDRShift");
	if (!armletData->egdrShift) return memErrNotEnoughSpace;
	armletData->sampleTab=GetRes(tblType, IDX_SAMPLETAB, "SampleTab");
	if (!armletData->sampleTab) return memErrNotEnoughSpace;
	armletData->sintab=GetRes(tblType, IDX_SINTAB, "SinTab");
	if (!armletData->sintab) return memErrNotEnoughSpace;
	armletData->resonanceLP=GetRes(tblType, IDX_RESONANCELP, "ResonanceLP");
	if (!armletData->resonanceLP) return memErrNotEnoughSpace;
	armletData->resonanceHP=GetRes(tblType, IDX_RESONANCEHP, "ResonanceHP");
	if (!armletData->resonanceHP) return memErrNotEnoughSpace;

	for (i=0; i<12; i++) {
		armletData->errors1541[i]=GetRes(strType, IDX_ERRORS1541+i,
			"Errors1541");
		if (!armletData->errors1541[i])
			return memErrNotEnoughSpace;
	}
	// Snapshot buffers
	if (!(armletData->snapshot1=NewSwapped(SNAPSHOT_SIZE_1)))
		return memErrNotEnoughSpace;
	if (!(armletData->snapshot2=NewSwapped(SNAPSHOT_SIZE_2)))
		return memErrNotEnoughSpace;
	
	// Set up prefs
	// Normally we want the Armlet to set it's prefs based on the saved ones
	armletData->cmd |= CMD68K_SETPREFS;
	armletData->prefs=CopyRes2(prefsType, IDX_PREFS, "Preferences", 1);
	if (!armletData->prefs) {
		SysFatalAlert("No prefs in Frodo.prc?!");
			return error;
	}
	// store a normal-endian pointer in prefs
	prefs=(struct PrefsState *) swap32(armletData->prefs);
	if (IsNuked(prefs) || crashed) {
		// Tell armlet to give us the default prefs if first time or Frodo crashed
		armletData->cmd |=   CMD68K_GETPREFS;
		armletData->cmd &= ~(CMD68K_SETPREFS);
	} else {
		// Only set the key mask if we have prefs, or all keys will be blocked!
		SetKeyMask();
	}
	if (IsNuked(prefs))
		// 'Nuked' prefs is a good indicator we are running Frodo for the first time.
		firstTime=true;
	
	//if (prefs->Sound) canSound=1;

	// Preload joystick bitmaps
	for (i=0; i<9; i++)
		if (joyHdl[i]=DmGetResource(bitmapType, i+32))
			joyPtr[i]=MemHandleLock(joyHdl[i]);

	return 0;
}

FreeArmletData()
{
	int i;
	
	if (!armletData) return;

	// Free up everything...
	// Things that where allocated by MemPtrNew or CopyRes
	// should get passed to MemPtrFree,
	// things that where created by GetResource should have their handles
	// unlocked.
	// It is SUPREMELY important that allocs and frees match!

	FreeSwapped(armletData->ram1);
	FreeSwapped(armletData->ram2);
	UncopyRes(armletData->kernalROM);
	UngetRes(armletData->basicROM);
	UngetRes(armletData->charROM);
	FreeSwapped(armletData->color);
	FreeSwapped(armletData->ram1541);
	UncopyRes(armletData->rom1541);
	UngetRes(armletData->expTable);
	UngetRes(armletData->multiExpTable);
	UngetRes(armletData->textColorTable);
	UngetRes(armletData->triTable);
	UngetRes(armletData->triSawTable);
	UngetRes(armletData->triRectTable);
	UngetRes(armletData->sawRectTable);
	UngetRes(armletData->triSawRectTable);
	UngetRes(armletData->egTable);
	UngetRes(armletData->egdrShift);
	UngetRes(armletData->sampleTab);
	UngetRes(armletData->sintab);
	UngetRes(armletData->resonanceLP);
	UngetRes(armletData->resonanceHP);

	for (i=0; i<12; i++)
		UngetRes(armletData->errors1541[i]);

	FreeSwapped(armletData->snapshot1);
	FreeSwapped(armletData->snapshot2);
	
	UncopyRes(armletData->prefs);

	MemPtrFree(armletData);
	armletData=NULL;
	
	for (i=0; i<9; i++) {
		if (joyPtr[i])
			MemHandleUnlock(joyHdl[i]);
		if (joyHdl[i])
			DmReleaseResource(joyHdl[i]);
	}
}

// Memory and resources 

// MemPtrNew with byteswapping
MemPtr NewSwapped(UInt32 n)
{
	MemPtr p=MemPtrNew(n);
	return (MemPtr) swap32(p);
}
void FreeSwapped(MemPtr p)
{
	MemPtr p2=(MemPtr) swap32(p);
	if (p2) MemPtrFree(p2);
}

// Load a resource and complain if it can't be found
MemPtr GetRes(DmResType t, UInt32 idx, char *name)
{
	MemHandle h;
	MemPtr p;

	h=DmGetResource(t, idx);
	if (!h) {
		StrPrintF(msg, "Couldn't find resource '%s' (%ld)", name, idx);
		SysFatalAlert(msg);
		return NULL;
	}
	p=MemHandleLock(h);
	if (!p) {
		DmReleaseResource(h);
		return NULL;
	}
	return (MemPtr) swap32(p);
}
void UngetRes(MemPtr p)
{
	MemHandle h;
	MemPtr p2;

	if (!p) return;

	p2=(MemPtr) swap32(p);

	h=MemPtrRecoverHandle(p2);
	if (!h) {
		StrPrintF(msg, "Couldn't recover handle for 0x%lx!", p2);
		if (!exiting) SysFatalAlert(msg);
	}
	MemHandleUnlock(h);
	DmReleaseResource(h);
}

// Same as above, but copy the resource to dynamic (writable) memory
// (which also means that you can release the resource immediately)

// Loud version
MemPtr CopyRes(DmResType t, UInt32 idx, char *name)
{
	return CopyRes2(t, idx, name, 0);
}
// Variable volume
MemPtr CopyRes2(DmResType t, UInt32 idx, char *name, int quiet)
{
	MemHandle h;
	MemPtr src, dst;

	h=DmGetResource(t, idx);
	if (!h) {
		if (!quiet) {
			StrPrintF(msg, "Couldn't find resource '%s'", name);
			SysFatalAlert(msg);
		}
		return NULL;
	}
	// Allocate memory, copy resource
	dst = MemPtrNew(MemHandleSize(h));
	if (!dst) {
		DmReleaseResource(h);
		if (!quiet) {
			StrPrintF(msg, "Out of memory loading '%s'", name);
			SysFatalAlert(msg);
		}
		return NULL;
	}
	src = MemHandleLock(h);
	MemMove(dst, src, MemHandleSize(h));
	MemHandleUnlock(h);
	DmReleaseResource(h);

	return (MemPtr) swap32(dst);
}
void UncopyRes(void *p)
{
	MemPtr p2=(MemPtr) swap32(p);

	if (p2) MemPtrFree(p2);
}

void SaveSnapshot(char *name, Int32 i)
{
	MemHandle h1,h2;
	MemPtr p1,p2;
	Boolean ok=false;
	UInt16 idx, idx2;
	UInt16 n;
	
	armletData->cmd &= ~(CMD68K_SAVE);
	if (!mainDBRef) return;
	
	n=DmNumRecords(mainDBRef);
	idx=i;
	if (i>=0 && i<n) {
		// Overwrite
		//SysFatalAlert("Old snapshot");
		if (h1=DmGetRecord(mainDBRef, idx))
			h2=DmGetRecord(mainDBRef, idx+1);
	} else {
		// New
		//SysFatalAlert("New snapshot");
		idx=dmMaxRecordIndex;
		if (h1=DmNewRecord(mainDBRef, &idx, SNAPSHOT_SIZE_1+36)) {
			idx2=idx+1;
			h2=DmNewRecord(mainDBRef, &idx2, SNAPSHOT_SIZE_2+36);
		}
	}
	if (h1) {
	    if (h2) {
		p1=MemHandleLock(h1);
		if (p1) {
			p2=MemHandleLock(h2);
			if (p2) {
				//SysFatalAlert("Writing snapshot");
				DmWrite(p1, 0, "SNAP", 4);
				DmWrite(p1, 4, name, 32);
				DmWrite(p1, 36,
					(MemPtr)swap32(armletData->snapshot1),
					SNAPSHOT_SIZE_1);
				DmWrite(p2, 0, "SNP2", 4);
				DmWrite(p2, 4, name, 32);
				DmWrite(p2, 36,
					(MemPtr)swap32(armletData->snapshot2), 
					SNAPSHOT_SIZE_2);
				//SysFatalAlert("Wrote snapshot");
				ok=true;
				MemHandleUnlock(h2);
			}
			MemHandleUnlock(h1);
		}
		DmReleaseRecord(mainDBRef, idx+1, ok);
	    }
	    DmReleaseRecord(mainDBRef, idx, ok);
	}
	if (!ok)
		if (!exiting) SysFatalAlert("Couldn't write snapshot!");
	
}

// Load a snapshot, d64 or t64 from the main db
// Go by id if fnam is null
// lastName is set on succesfull load, and is used to suggest a file name on quicksave
char lastName[32]="Unnamed";
void Load(DbFile *dbf, UInt16 id)
{
	MemHandle h1,h2;
	UInt8 *p1=NULL, *p2;
	char name[32];
	Boolean ok=false, nuked=false;
	UInt16 i;
	Int32 hit=-1;
	Boolean snap=false;
	DbFile main;
	Err err;

	//StrPrintF(msg, "mainDB is 0x%lx", mainDBRef);
	//SysFatalAlert(msg);
	if (!dbf) {
		dbf=&main;
		DbFileFromDbRef(mainDBRef, dbf);
	}
	//StrPrintF(msg, "Load from %d, 0x%lx, %d", dbf->external, dbf->u.db, id);
	//SysFatalAlert(msg);
	if (!dbf->u.db) return;

	// Find by index
	//StrPrintF(msg, "Load by index %d", id);
	//SysFatalAlert(msg);
	// if (h1=DmGetRecord(mainDBRef, id)) {
	//StrPrintF(msg, "Try idx %lx %d", dbf->u.db, id);
	//SysFatalAlert(msg);
	//  DbFileGetRecord(&dbFile[i], j, &h);
	err = DbFileGetRecord(dbf, id, &h1);
	//StrPrintF(msg, "Did try idx %lx %d", dbf->u.db, id);
	//SysFatalAlert(msg);

	if (h1) {
		if (p1=MemHandleLock(h1)) {
			if (!MemCmp(p1, "SNAP", 4) || !MemCmp(p1, "SPLT", 4)) {
				hit=id;
				if (!MemCmp(p1, "SNAP", 4))
					snap=true;
			} else {
				StrPrintF(msg, "Invalid type \"%c%c%c%c\"", 
					p1[0],p1[1],p1[2],p1[3]);
				SysFatalAlert(msg);
				p1=NULL;
				MemHandleUnlock(h1);
			}
		} else {
			StrPrintF(msg, "Couldn't lock handle");
			SysFatalAlert(msg);
			//DmReleaseRecord(mainDBRef, id, false);
			DbFileReleaseRecord(dbf, id, h1, false);
		}
	}
	
	if (hit<0) {
		return;
	}

	// File found, and locked in p1.
	if (!MemCmp(p1, "SNAP", 4)) {
		// Snapshots
		// h2=DmGetRecord(mainDBRef, hit+1);
		err = DbFileGetRecord(dbf, hit+1, &h2);
		if (h2) {
			p2=MemHandleLock(h2);
			if (p2) {
				if (IsNuked(p1+36) || IsNuked(p2+36))
					nuked=true;
				else if (!MemCmp(p2, "SNP2", 4) &&
				    !StrNCompare(p1+4, p2+4, 32)) {
				        MemMove((MemPtr)swap32(armletData->snapshot1), 
							p1+36, SNAPSHOT_SIZE_1);
					MemMove((MemPtr)swap32(armletData->snapshot2), 
							p2+36, SNAPSHOT_SIZE_2);
					ok=true;
					StrNCopy(lastName, p1+4, 32);
					lastName[31]='\0';
				}
				MemHandleUnlock(h2);
			}
			//DmReleaseRecord(mainDBRef, hit+1, false);
			DbFileReleaseRecord(dbf, hit+1, h2, false);
		}
		MemHandleUnlock(h1);
		// DmReleaseRecord(mainDBRef, hit, false);
		DbFileReleaseRecord(dbf, hit, h1, false);
		if (ok) {
			armletData->cmd |= CMD68K_LOAD;
			//StrPrintF(msg, "cmd=%lx", armletData->cmd);
			//SysFatalAlert(msg);
		} else {
			armletData->cmd &= ~(CMD68K_LOAD);
			if (!(nuked && id==0))
				SysFatalAlert("Failed to load snapshot");
		}
	} else if (!MemCmp(p1, "SPLT", 4)) {
		// SplitFile .d64 or .t64

/* Current layout:
	00 'SPLT' (first part) || 'PART' (other parts)
	04 char name[32]     // must have terminating 0, so max 31 letters.
	36 int32 totalParts
	40 int32 part
	44 int32 size        // File size
	48 int32 partSize    // Part size. All parts must be the same size (except the last)
	52 char data[partSize]
*/
		int nParts;
		//UInt32 DBGoffs=0x1df00L;
		Close(swap32(armletData->disk[0]));
		armletData->disk[0]=NULL;
		MemSet(&(disk[0]), sizeof(FILE), 0x0);
		MemMove(disk[0].name, p1+4, 32);
		disk[0].part[0].p=p1;

		nParts=*((UInt32 *)(p1+36));
		disk[0].size=*((UInt32 *)(p1+44));
		disk[0].partSize=*((UInt32 *)(p1+48));
		disk[0].part[0].handle=(UInt32)h1;
		if (dbf->external) {
			disk[0].part[0].id=0xffff;
		} else {
			disk[0].part[0].dbf=(UInt32)dbf;
			disk[0].part[0].id=hit;
		}
		ok=true;
		//StrPrintF(msg, "Parts: %d PartSize: %ld Size: %ld",
		//	nParts, disk[0].partSize, disk[0].size);
		//SysFatalAlert(msg);
		
		for (i=0; i<nParts; i++) {
			int n, p;
			UInt32 s, ps;
			if (i>0) {
				// next part
				//if (!(disk[0].part[i].handle=(UInt32)h1=DmGetRecord(mainDBRef, id+i))) {
				err = DbFileGetRecord(dbf, id+i,
						(MemHandle *) &(disk[0].part[i].handle));
				h1=(MemHandle) disk[0].part[i].handle;
				if (err) {
					ok=false;
					break;
				}
				if (dbf->external)
					disk[0].part[i].id=0xffff;
				else
					disk[0].part[i].id=id+i;
					
				if (!(disk[0].part[i].p=p1=MemHandleLock(h1))) {
					ok=false;
					break;
				}
				
				/*if (DBGoffs>disk[0].partSize) {
					DBGoffs-=disk[0].partSize;
				} else {
					sprintf(msg, "At 0x1df00: 0x%02x%02x",
						p1[DBGoffs],
						p1[DBGoffs+1]);
					SysFatalAlert(msg);
				}*/
				if (MemCmp(p1, "PART", 4)) {
					ok=false;
					break;
				}
			}
			n=*((UInt32 *)(p1+36));
			p=*((UInt32 *)(p1+40));
			s=*((UInt32 *)(p1+44));
			ps=*((UInt32 *)(p1+48));
			
			if (n!=nParts || p!=i || s!=disk[0].size ||
			    ((i<p-1) && ps!=disk[0].partSize)) {
				ok=false;
				//StrPrintF(msg, "Couldn't load: %d<>%d or %d<>%d or %ld<>%ld or %ld<>%ld",
				//	n,nParts, p,i, s,disk[0].size, ps,disk[0].partSize);
				//SysFatalAlert(msg);
				break;
			}
			disk[0].part[i].p=p1+52;
			if (i==nParts-1) {
				StrNCopy(lastName, p1+4, 32);
				lastName[31]='\0';
			}
			//StrPrintF(msg, "Part %d at %lxd", i, disk[0].part[i].p); 
			//SysFatalAlert(msg);
		}
		if (ok) {
			disk[0].size=swap32(disk[0].size);
			disk[0].partSize=swap32(disk[0].partSize);
			for (i=0; i<nParts; i++)
				disk[0].part[i].p=(UInt8 *)
					swap32(disk[0].part[i].p);
			armletData->disk[0]=(FILE *) swap32(&(disk[0]));
			if (!StrCompare(disk[0].name+StrLen(disk[0].name)-4, ".d64"))
				prefs->DriveType[0]=swap32(DRVTYPE_D64);
			else
				prefs->DriveType[0]=swap32(DRVTYPE_T64);
			MemMove(prefs->DrivePath[0], disk[0].name, 32);

			armletData->cmd |= CMD68K_SETPREFS;
		} else {
			armletData->cmd &= ~(CMD68K_SETPREFS);
			StrPrintF(msg, "Failed to load disk image \"%s\"", disk[0].name);
			SysFatalAlert(msg);
		}
	} else {
		StrPrintF(msg, "Unknown file type \"%c%c%c%c\"",
			p1[0],p1[1],p1[2],p1[3]);
		SysFatalAlert(msg);
		MemHandleUnlock(h1);
		//DmReleaseRecord(mainDBRef, hit, false);
		DbFileReleaseRecord(dbf, hit, h1, false);
	}
	//SysFatalAlert("Load ends");
}
Delete(DbFile *dbf, UInt16 idx)
{
	MemHandle h;
	MemPtr p;
	UInt16 i, n=DmNumRecords(mainDBRef);
	Boolean stop=false;
	DbFile main;
	Err err;
	
	if (!dbf) {
		dbf=&main;
		DbFileFromDbRef(mainDBRef, dbf);
	}

	// Sanity checks
	if (dbf->external) return;
	if (!dbf->u.db) return;
	if (dbf->u.db==mainDBRef && idx<2) return;
	if (idx>=n) return;

	for (i=idx; i<n && !stop; i++) {
		//if (h=DmGetRecord(mainDBRef, i)) {
		err = DbFileGetRecord(dbf, i, &h);
		if (h) {
			if (p=MemHandleLock(h)) {
				if (i==idx) {
					if (StrNCompare(p, "SNAP", 4) && StrNCompare(p, "SPLT", 4))
						stop=1;
				} else if (!StrNCompare(p, "SNAP", 4) || !StrNCompare(p, "SPLT", 4))
					stop=1;
				MemHandleUnlock(h);
			} else
				stop=true;
			//DmReleaseRecord(mainDBRef, i, false);
			DbFileReleaseRecord(dbf, i, h, false);
		} else
			stop=true;
		if (!stop) DmDeleteRecord(dbf->u.db, i);
	}
}
// Check for orphaned parts and remove them from the db
/*
FixDB()
{
	MemHandle h;
	MemPtr p;
	UInt16 i, n=DmNumRecords(mainDBRef);
	Boolean del;
	char name[32]="FIRST", *expect="SNAP";

	for (i=2; i<n; i++) {
		del=false;
		if (h=DmGetRecord(mainDBRef, i)) {
			if (p=MemHandleLock(h)) {
				if (!StrNCompare(p, "SNAP", 4)) {
					StrNCopy(name, p+4, 32);
					name[31]='\0';
					expect="SNP2";
				} else if (!StrNCompare(p, "SPLT", 4)) {
					StrNCopy(name, p+4, 32);
					name[31]='\0';
					expect="PART";
				} else if (StrNCompare(p, expect, 4) <<
					   StrNCompare(p+4, name, 32)) {
					StrPrintF(msg, "Expected %s%s, "\
						", not %.36s!",
						expect, name, p);
					SysFatalAlert(msg);
					del=true;
				}
				MemHandleUnlock(h);
			} else {
				StrPrintF(msg, "Couldn't lock record %d", i);
				SysFatalAlert(msg);
			}
			DmReleaseRecord(mainDBRef, i, false);
		} else {
			// It's probably deleted, or in use
			//StrPrintF(msg, "Couldn't get record %d", i);
			//SysFatalAlert(msg);
		}
		if (del) {
			if (DmDeleteRecord(mainDBRef, i))
				SysFatalAlert("Couldn't delete!");
			else {
				n--;
				i--;
			}
		}
	}
}
*/

// Close a splitfile
Close(FILE *f)
{
	int i;

	if (!f) return;
	
	//StrPrintF(msg, "Close file %lxd", f);
	//if (!exiting) SysFatalAlert(msg);
	
	for (i=0; i<MAX_PARTS; i++) {
		/*StrPrintF(msg, "Part %d handle=%lx", i, f->part[i].handle);
		SysFatalAlert(msg);
		*/
		if (f->part[i].handle) {
			/*StrPrintF(msg, "Part %d pointer=%lx", i, f->part[i].pointer);
			SysFatalAlert(msg);*/
			if (f->part[i].p)
				MemHandleUnlock((MemHandle)f->part[i].handle);
			/*StrPrintF(msg, "Part %d id=%ld", i, f->part[i].id);
			SysFatalAlert(msg);*/
			if (f->part[i].id!=0xffff)
				DmReleaseRecord(mainDBRef, f->part[i].id, false);
			else
				MemHandleFree((MemHandle)f->part[i].handle);
		}
		f->part[i].handle=NULL;
		f->part[i].p=NULL;
		f->part[i].dbf=NULL;
	}
}

// Set correct key mask
SetKeyMask()
{
	UInt32 mask=0;
	
	if (!prefs) return;
	
	mask=prefs->KeyUp | prefs->KeyDown | prefs->KeyLeft | prefs->KeyRight | prefs->KeyFire;
	mask=swap32(~mask);
	//StrPrintF(msg, "Setting key mask 0x%lx", mask);
	//SysFatalAlert(msg);
	KeySetMask(mask);
}


UInt32 ChangeScreenMode(UInt32 sm, Boolean first, Boolean defaults)
{
	UInt32 w,h,d;
	Boolean col=true;
	UInt32 oldsm;
	Boolean buffer = (sm==SCREENMODE_256_BUFFERED);
	UInt32 error;
	UInt16 i16;
	
	oldsm=swap32(prefs->ScreenMode);
	// Check for unreasonable screen modes
	if (!is320 ||!can256) {
		buffer=true;
		sm=SCREENMODE_256_BUFFERED;
	}
	
	if (sm==SCREENMODE_16 || sm==SCREENMODE_16ALT)
		d=4;
	else
		d=8;
	
	if (d==4 && !can16) {
		d==8;
		sm=SCREENMODE_256;
	}
	if (d==8 && !buffer && !can256) {
		sm=SCREENMODE_256_BUFFERED;
		buffer=true;
	}
	
	if (sm==swap32(prefs->ScreenMode) && !first && !defaults)
		return 0;
			
	if (buffer) {
		//int i;
		// Draw to a bitmap
		// Use the default screen mode in this case
		if (!IsNuked(prefs)) {
			WinScreenMode(winScreenModeSetToDefaults, NULL,NULL,NULL,NULL);
		}
		// Wrong size or no 16-color mode. Render to bitmap.
		/*d=8;
		if (error=WinScreenMode(winScreenModeSet, &w, &h, &d, &col)) {
			StrPrintF(msg, "Couldn't change screen mode (%ld)",
				error);
			SysFatalAlert(msg);
			return error;
		}*/
		//for (i=0; i<16; i++) {
		  if (error=WinPalette(winPaletteSet, 0 /* i*16 */, 16, colorTable.pal)) {
			StrPrintF(msg, "Couldn't change palette (0x%lx)", error);
			SysFatalAlert(msg);
			return error;
		  }
		//} 
		//bmp=BmpCreate(320, 272, 4, (ColorTableType *) &colorTable, &i16);
		//bmp=BmpCreate(320, 272, 4, NULL, &i16);
		bmp=BmpCreate(320, 272, 8, NULL, &i16);
		if (!bmp) {
			StrPrintF(msg, "Error %d in BmpCreate()", i16);
			SysFatalAlert(msg);
			return 1;
		}
		bmpv3=BmpCreateBitmapV3(bmp, kDensityDouble, BmpGetBits(bmp), NULL);
		if (!bmpv3) {
			SysFatalAlert("Couldn't create high-res Bitmap");
			return 1;
		}
		bmpBits=BmpGetBits((BitmapType *)bmpv3);
		//bmpBits=BmpGetBits((BitmapType *)bmp);
		
		if (((UInt32)bmpBits) & 3) {
			SysFatalAlert("bmpBits not aligned!");
			return 1;
		}
		if (!bmpBits) {
			SysFatalAlert("Couldn't get bitmap bits");
			return 1;
		}
		armletData->displayBuf = (UInt8 *) swap32(bmpBits);
		armletData->displayModulo = swap32(320);
	} else  {
		// Direct memory access
		// If we crashed, or this is the first time Frodo is run
		// don't do any screen mode changes until we get the default prefs
		if (!IsNuked(prefs)) {	
		    if (error=WinScreenMode(winScreenModeSet, &w, &h, &d, &col)) {
			StrPrintF(msg, "Couldn't change screen mode (%ld)",
				error);
			SysFatalAlert(msg);
			return error;
		    }
		}
		if (error=WinPalette(winPaletteSet, 0, 16, colorTable.pal)) {
			StrPrintF(msg, "Couldn't change palette (0x%lx)", error);
			SysFatalAlert(msg);
			return error;
		}
		screenBits=(UInt8 *) BmpGetBits(WinGetBitmap(WinGetDrawWindow()));
		if (!screenBits) {
			StrPrintF(msg, "Couldn't get screen bits");
			SysFatalAlert(msg);
			return 1;
		}
		if (bmpBits) {
			BmpDelete(bmp);
			BmpDelete((BitmapPtr)bmpv3);
			bmp=NULL;
			bmpv3=NULL;
		}
		armletData->displayBuf = screenBits;
		if (d==4)
			armletData->displayModulo = swap32(screenW/2);
		else
			armletData->displayModulo = swap32(screenW);
			
		// There's an off chance that the device will report double
		// density but it does in fact have a width other than 320 pixels.
		// <320 was tackled in CheckPrerequisites, >320 right here.
		if (screenW>320) {
			// ... and it doesn't work at all on eg. UX50 480x320
			// becuase it LIE's about the true width.
			// With no way to get the true width except device ID.
			// Let them use "compatible mode".
			UInt16 offs=screenW-320;
			UInt16 scroll;
			
			if (d==4) offs/=2;
			scroll=offs&0xfffc; // Center, keep it aligned
			
			armletData->displayBuf+=scroll;
		}
		armletData->displayBuf = (UInt8 *) swap32(armletData->displayBuf);
	}
	prefs->ScreenMode=swap32(sm);
	if (crashed || !first) SavePrefs();
	return 0;
}

// Fix up displayBuf and soundBuf
UInt32 InitGraphicsSound()
{
	UInt32 error;
	
	if (error=ChangeScreenMode(swap32(prefs->ScreenMode), true, false))
		return error;

	// Now the sound
	// Set the buffer even if we won't use it (just in case...)
	if (!(soundStruct.soundBuf=MemPtrNew(SOUND_BUFSIZE*2))) {
		SysFatalAlert("Couldn't allocate sound buffer! Turning off sound.");
		canSound=false;
	}
	armletData->soundBuf = (UInt16 *) swap32(&(armletBuf[0]));

	if (canSound)
		StartSound();
	
	return 0;
}
StartSound()
{
	Err error;
	
	if (sndStream) {
		// Stream exists. Assume it was paused.
		SndStreamPause(sndStream, false);
		SndStreamSetVolume(sndStream, swap32(prefs->Volume));
		return 0;
	}
	// ..Extended is 5.1? 5.2? Not 5.0 at any rate
	/*error = SndStreamCreateExtended(&sndStream, sndOutput, sndFormatPCM, 44100,
		sndInt16Little, sndMono, SoundCallback, armletData, 882, false);
	*/
	/*
	error = SndStreamCreateExtended(&sndStream, sndOutput, 0, 44100,
		sndInt16Little, sndMono, SoundCallback, armletData, 882, false);
	*/
	// SoundCallback will be called whenever the hw needs samples.
	// We have no way of knowing how many samples will be demanded,
	// and must deliver them STAT. Oh joy.
	error = SndStreamCreate(&sndStream, sndOutput, 44100,
		sndInt16Little, sndMono, SoundCallback, &soundStruct,
			FRAME_SAMPLES, false);
	if (error) {
		SysFatalAlert("Couldn't start sound");
		sndStream=NULL;
	} else {
		//if (prefs->Volume) {
		// Prefs mightn't'v been loaded yet! wait 1f before starting the stream?
			SndStreamStart(sndStream);
			SndStreamSetVolume(sndStream, swap32(prefs->Volume));
		//}
	}
	return 0;
}
StopSound()
{
	if (sndStream)
		SndStreamPause(sndStream, true);

	return 0;
}

// Shut of audio, release bitmaps, reset screen mode
FreeGraphicsSound()
{
	if (!armlet) return;

	WinScreenMode(winScreenModeSetToDefaults, NULL,NULL,NULL,NULL);
	if (bmpBits) {
		BmpDelete(bmp);
		BmpDelete((BitmapType *) bmpv3);
	}
	if (sndStream) SndStreamDelete(sndStream);
	if (soundStruct.soundBuf)
		MemPtrFree(soundStruct.soundBuf);
}
/* Explanation: (See also VBlank) the SoundStruct contains a large(ish)
   ringbuffer of samples. SampleOffset is the index of the next sample that
   hasn't been output. SamplesQueued inidicates how many samples have been
   produced by the armlet beyond that point.
   SamplesQueued can never grow larger than the size of the buffer
   (if it does the data is thrown away rather than being queued.)
   The parameter 'size' is how many samples we must write to bufferP.
   Since I just fill any underflow with zeros, the net effect of wildly
   fluctuating frame rates should be that the latency of the stream 
   automatically goes up (amidst horrible clicking).
   After a while the audio should be nice again (but high latency).
   This will ONLY work if the buffer size is larger than the maximum lag!
   Currently 4kB=0.1 s, kinda low.
   In case of pause, modal dialogs or the like the ringbuffer will quickly be
   emptied, and the whole process starts over when the emulator starts running
   normally again.
   If the Palm simply isn't fast enough to run at an average speed of 50Hz
   (even when dropping 90% of the frames) the sound WILL be shot to hell.
   That's what the 'Mute' option is for.
*/
Err SoundCallback(void *userDataP, SndStreamRef stream,
                  void *bufferP, UInt32 size)
{
	struct SoundStruct *ss=(struct SoundStruct *) userDataP;
	UInt16 *p=(UInt16 *)bufferP;
	UInt32 nsmpl=size, nempty=0;
	UInt32 i, j;

	// Feedback noise when adjusting volume
	if (ss->makeNoise) {
		// ~800 Hz square wave, 64 samples per wave
		i=0;
		// Finish previous half-cycle
		j=32-(ss->noiseOffs&31);
		if (j>size) j=size;
		if (ss->noiseOffs&32)
			MemSet(p, j*2, 0x1f);
		else
			MemSet(p, j*2, 0xe0);
		i+=j;
		ss->noiseOffs = (ss->noiseOffs+j)&63;
		
		// Produce square waves, the last will be truncated.
		while (i<size) {
			j=32;
			if (i+j > size) j=size-i;
			// 1/4 max volume, like a single SID voice.
			if (ss->noiseOffs&32)
				MemSet(p+i, j*2, 0x1f);
			else
				MemSet(p+i, j*2, 0xe0);
			i+=j;
			ss->noiseOffs = (ss->noiseOffs+j)&63;
		}
		return errNone;
	}
	
	// The real deal
	if (size > ss->samplesQueued) {
		// Underflow, fill with 0 
		nempty = size-ss->samplesQueued;
		nsmpl = ss->samplesQueued;
	}
	if (ss->sampleOffset+nsmpl > SOUND_BUFSIZE) {
		MemMove(p, ss->soundBuf+ss->sampleOffset,
			(SOUND_BUFSIZE - ss->sampleOffset)*2);
		MemMove(p+SOUND_BUFSIZE - ss->sampleOffset, ss->soundBuf, 
			(ss->sampleOffset+nsmpl - SOUND_BUFSIZE)*2);
		ss->sampleOffset += nsmpl;
		ss->sampleOffset -= SOUND_BUFSIZE;
	} else {
		MemMove(p, ss->soundBuf+ss->sampleOffset, nsmpl*2);
		ss->sampleOffset += nsmpl;
	}
	// It seems an idle SID produces 0xe000
	MemSet(p+nsmpl, nempty*2, 0xdf);
	ss->samplesQueued -= nsmpl;
	ss->underflow += size;
	return errNone;
}

#define DB_NAME "Frodo main database"
InitDB()
{
	Err e;
	UInt16 at;
	UInt16 card;
	LocalID id;
	DmOpenRef dataDBRef;
	Boolean new=false;
	MemHandle h;
	MemPtr p;
	DmSearchStateType search;
	UInt16 i, attr;
	
	mainDB=DmFindDatabase(0, DB_NAME);
	if (!mainDB) {
		e=DmCreateDatabase(0, DB_NAME, appSignature, mainType, false);
		if (e) {
			StrPrintF(msg, "Error creating database: %d", e);
			SysFatalAlert(msg);
			return 0;
		}
		// TODO: backup flag!
		new=true;
	}
	mainDB=DmFindDatabase(0, DB_NAME);
	if (!mainDB) {
		StrPrintF(msg, "Silent error creating database??");
		SysFatalAlert(msg);
		return 0;
	}
	attr=dmHdrAttrBackup | dmHdrAttrBundle;
	DmSetDatabaseInfo(0, mainDB, NULL, &attr, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL);

	mainDBRef=DmOpenDatabase(0, mainDB, dmModeReadWrite);
	if (!mainDBRef) {
		StrPrintF(msg, "Error opening database");
		SysFatalAlert(msg);
		return 0;
	}
	if (new) {
		at=0;
		h=DmNewRecord(mainDBRef, &at, SNAPSHOT_SIZE_1+36);
		if (!h) {
			StrPrintF(msg, "Couldn't create snapshot 1: %d", DmGetLastErr());
			SysFatalAlert(msg);
			DmCloseDatabase(mainDBRef);
			DmDeleteDatabase(0, mainDB);
			mainDB=0;
			mainDBRef=NULL;
			return 0;
		}
		p=MemHandleLock(h);
		DmWrite(p, 0, "SNAP", 4);
		DmWrite(p, 4, "Nuked snapshot\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 32);
		DmWrite(p, 36, "NUKE", 4); 
		MemHandleUnlock(h);
		DmReleaseRecord(mainDBRef, at, true);
		at=1;
		h=DmNewRecord(mainDBRef, &at, SNAPSHOT_SIZE_2+36);
		if (!h) {
			StrPrintF(msg, "Couldn't create snapshot 2: %d", DmGetLastErr());
			SysFatalAlert(msg);
			DmCloseDatabase(mainDBRef);
			DmDeleteDatabase(0, mainDB);
			mainDB=0;
			mainDBRef=NULL;
			return 0;
		}
		p=MemHandleLock(h);
		DmWrite(p, 0, "SNP2", 4);
		DmWrite(p, 4, "Nuked snapshot\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 32);
		DmWrite(p, 36, "NUKE", 4); 
		MemHandleUnlock(h);
		DmReleaseRecord(mainDBRef, at, true);
	}
	// Database is open. Look for other Frodo PDB's and merge them into the main db.

	//FixDB();
	if (DmGetNextDatabaseByTypeCreator(
			true, &search, dataType, appSignature, false, &card, &id)==errNone) {
		do {
			UInt16 n;
			//StrPrintF(msg, "Found some warez!");
			//SysFatalAlert(msg);
			dataDBRef=DmOpenDatabase(card, id, dmModeReadWrite);
			if (!dataDBRef) {
				StrPrintF(msg, "Can't open secondary database");
				SysFatalAlert(msg);
				continue;
			}
			n=DmNumRecords(dataDBRef);
			for (i=0; i<n; i++) {
				if (e=DmDetachRecord(dataDBRef, 0, &h)) {
					StrPrintF(msg, "Detach failed for %d: %d", at, e);
					SysFatalAlert(msg);
					continue;
				}
				at=dmMaxRecordIndex;
				if (e=DmAttachRecord(mainDBRef, &at, h, NULL)) {
					StrPrintF(msg, "Ooops, lost some data: %d", e);
					SysFatalAlert(msg);
				}
				//StrPrintF(msg, "Move ext %d to int %d", i, at);
				//SysFatalAlert(msg);
			}
			DmCloseDatabase(dataDBRef);
			DmDeleteDatabase(card, id);
		} while (DmGetNextDatabaseByTypeCreator(
			false, &search, dataType, appSignature, false, &card, &id)==errNone);
	}
	
	// We'll do this everytime just in case
	VFSRegisterDefaultDirectory(".d64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".t64", expMediaType_Any, frodoDir);
	// Seems to be case insensitive, but just to be safe
	VFSRegisterDefaultDirectory(".D64", expMediaType_Any, frodoDir);
	VFSRegisterDefaultDirectory(".T64", expMediaType_Any, frodoDir);
	// Stand-alone snapshots are not supported yet
	VFSRegisterDefaultDirectory(".snp", expMediaType_Any, frodoDir);

	// If we crashed last time the snapshot is not to be trusted. "Safe Mode"...
	if (!new && !crashed) {
		Load(NULL, 0);
	}
	return 0;
}
FreeDB()
{
	int i;
	
	for (i=0; i<4; i++)
		Close((FILE *) swap32(armletData->disk[i]));
	
	DmCloseDatabase(mainDBRef);
}

// Finally, go ARM!
UInt32 RunArmlet()
{
	UInt32 error;
	
	//SysFatalAlert("3");
	FrmGotoForm(MainForm);
	//SysFatalAlert("4");
	//DumpArmlet();
#ifdef DEBUG_GUI

	// Pretend armlet is running
	while (!armletData->errCode)
		EventLoop();
	/*
	while (!armletData->errCode) {
		VBlank68k(armletData);
	}
	*/
#else
	//StrPrintF(msg, "Armlet data is %lx", armletData);
	//SysFatalAlert(msg);
	error=PceNativeCall((MemPtr)armlet, armletData);
#endif
	return armletData->errCode;
}

  /*************************************************************************/
 /* Everything below this line is called via callbacks from the ARM side. */
/*************************************************************************/
// Incidentally, the clean separation of code above and below helps us
// avoid having to split the program into sections.

// Utility functions for the dialogs

int32 atoi(const char *s)
{
	int32 i=0, neg=0;
	
	if (!s) return 0;
	
	if (s[0]=='-' && s[1]>='0' && s[1]<='9') {
		neg=1;
		s++;
	}
	while (*s>='0' && *s<='9') {
		i=i*10+*s-'0';
		s++;
	}
	
	return neg? -i:i;
}
char *itoa(int i)
{
	static char a[10];

	StrPrintF(a, "%d", i);
	return a;
}

UInt32 sampledKey=0;
UInt32 SampleKey(UInt32 org)
{
	// Eat all keys (including power!).
	UInt32 mask=KeySetMask(0), choice;
	sampledKey=0;
	
	// The eventHandler sets sampledKey everytime the 'Sample' button is pressed.
	choice=ShowDialog(SampleKeyForm);
	
	// Back to normal.
	KeySetMask(mask);
	if (choice==SampleKeyOKButton)
		return sampledKey;
	else
		return org;
}

// Get a controls label or text
static const char *GetText(FormType *f, UInt16 id)
{
	static char s[51];
	UInt16 idx=FrmGetObjectIndex(f, id);
	FormObjectKind k;
	void *c;
	FieldType *fl;

	if (idx==frmInvalidObjectId)
		return NULL;

	c=FrmGetObjectPtr(f, idx);
	k=FrmGetObjectType(f, idx);
	switch (k) {
		case frmFieldObj:
			fl=(FieldType *)c;
			return FldGetTextPtr(fl);
			break;
		default:
			return CtlGetLabel((ControlType *) c);
	}
}
// Set a controls label or text to a string
static Boolean SetText(FormType *f, UInt16 id, char *s)
{
	UInt16 idx=FrmGetObjectIndex(f, id);
	FormObjectKind k;
	void *c;
	FieldType *fl;

	if (idx==frmInvalidObjectId) {
		StrPrintF(msg, "Error in SetText(0x%lx, %d, \"%s\")", f, id, s);
		SysFatalAlert(msg);
		return false;
	}

	c=FrmGetObjectPtr(f, idx);
	k=FrmGetObjectType(f, idx);
	switch (k) {
		case frmFieldObj:
			fl=(FieldType *)c;
			FldDelete(fl, 0, FldGetTextLength(fl)-1);
			FldInsert(fl, s, StrLen(s));
			break;
		default:
			FrmCopyLabel(f, id, s);
	}
	return true;
}
// Get the value of a control
static UInt16 GetValue(FormType *f, UInt16 id)
{
	UInt16 idx=FrmGetObjectIndex(f, id);
	void *c;

	if (idx==frmInvalidObjectId)
		return false;

	c=FrmGetObjectPtr(f, idx);
	return CtlGetValue((ControlType *)c);
}
// Set the value of a control
static Boolean SetValue(FormType *f, UInt16 id, UInt16 v)
{
	UInt16 idx=FrmGetObjectIndex(f, id);
	void *c;

	if (idx==frmInvalidObjectId)
		return false;

	c=FrmGetObjectPtr(f, idx);
	CtlSetValue((ControlType *)c, v);

	return true;
}

// Show one of the dialogs. Dialogs may or may not have a Init function, an
// Exit function and an event handler.
// Init is called after the form is created, but before it is displayed.
// Exit is called after the form is closed, but before it is destroyed.
static UInt16 ShowDialog(UInt16 id)
{
	FormType *f=FrmInitForm(id);
	UInt16 choice=0;
	void (*initP)(FormType *)=NULL, (*exitP)(FormType *, UInt16)=NULL;

	switch (id) {
		case GfxSoundForm:
			FrmSetEventHandler(f, HandleEventGfxSound);
			initP=InitGfxSoundDialog;
			exitP=ExitGfxSoundDialog;
			break;
		case JoystickForm:
			FrmSetEventHandler(f, HandleEventJoystick);
			initP=InitJoystickDialog;
			break;
		case ObscureForm:
			initP=InitObscureDialog;
			exitP=ExitObscureDialog;
			break;
		case SampleKeyForm:
			FrmSetEventHandler(f, HandleEventSampleKey);
			break;
	}
	if (initP) initP(f);
	choice=FrmDoDialog(f);
	if (exitP) exitP(f, choice);
	FrmDeleteForm(f);
	//FrmDrawForm(FrmGetActiveForm());
	return choice;
}
// Show the "Save as" requester, with an optional default file name
static char *ShowSaveAsDialog(char *s)
{
	FormType *f=FrmInitForm(SaveAsForm);
	UInt16 choice=0;
	static char b[32];
	const char *t;
	char *retval=NULL;
	
	if (s) {
		SetText(f, SaveAsNameField, s);
	}
	FrmSetFocus(f, FrmGetObjectIndex(f, SaveAsNameField));
	choice=FrmDoDialog(f);
	if (choice==SaveAsOKButton) {
		t=GetText(f, SaveAsNameField);
		if (t!=NULL) {
			StrNCopy(b, t, 32);
			b[31]='\0';
			retval=b;
		}
	}
	FrmDeleteForm(f);
	return retval;
}

UInt8 cmdLock=0, shiftLock=0, lowerCase=0;
UInt16 keyboardBmp=KeyboardBitmap;
// Sets the correct keyboard underlay
static void UpdateOnScreenKeyboard(FormType *f)
{
	UInt16 bmpid;
	UInt16 idx=FrmGetObjectIndex(f, keyboardBmp);

	if (idx==frmInvalidObjectId) {
		StrPrintF(msg, "UOK: Couldn't find object %d", keyboardBmp);
		SysFatalAlert(msg);
		return;
	}
	if (FrmRemoveObject(&f, idx)) {
		StrPrintF(msg, "UOK: Couldn't remove object %d", idx);
		SysFatalAlert(msg);
	}
	
	if (cmdLock) {
		keyboardBmp=KeyboardCmdBitmap;
	} else if (lowerCase) {
		if (shiftLock) {
			keyboardBmp=KeyboardLocaseShiftBitmap;
		} else {
			keyboardBmp=KeyboardLocaseBitmap;
		}
	} else if (shiftLock) {
		keyboardBmp=KeyboardShiftBitmap;
	} else {
		keyboardBmp=KeyboardBitmap;
	}

	if (!FrmNewBitmap(&f, keyboardBmp, keyboardBmp, 0, 136)) {
		StrPrintF(msg, "UOK: Couldn't create bitmap %d", keyboardBmp);
		SysFatalAlert(msg);
		return;
	}
	RepaintOnScreenKeyboard(f);
}
RepaintOnScreenKeyboard(FormType *f)
{
	FormBitmapType *b;
	RectangleType r;
	UInt16 idx=FrmGetObjectIndex(f, keyboardBmp);
	MemHandle m;
	BitmapType *bmp;
	
	if (idx==frmInvalidObjectId) {
		StrPrintF(msg, "Couldn't find object %d", keyboardBmp);
		SysFatalAlert(msg);
		return;
	}
	
	FrmGetObjectBounds(f, idx, &r);
	m=DmGetResource(bitmapType, keyboardBmp);
	if (m) {
		bmp=MemHandleLock(m);
		if (bmp) {
			WinDrawBitmap(bmp, r.topLeft.x, r.topLeft.y);
			MemHandleUnlock(m);
		} else {
			StrPrintF(msg, "Couldn't lock handle %lx", m);
			SysFatalAlert(msg);
		}
		DmReleaseResource(m);
	} else {
		StrPrintF(msg, "Couldn't find resource %d", keyboardBmp);
		SysFatalAlert(msg);
	}
	return;
}

UInt16 joyBmp=0;
UInt32 oldJoy=0xffffffffL;
static void UpdateOnScreenJoystick()
{
	UInt16 bmpid;
	UInt16 idx;
	UInt32 joy=swap32(armletData->joystick);
	RectangleType r;
	MemHandle m;
	BitmapType *bmp;
	FormPtr f;

	if (joy==oldJoy) return;
	
	oldJoy=joy;
	
	f=FrmGetActiveForm();
	if (FrmGetFormId(f)!=MainForm) return;

	if (joy & JOYUP) {
		if (joy & JOYLEFT)
			joyBmp=40;
		else if (joy & JOYRIGHT)
			joyBmp=34;
		else
			joyBmp=33;
	} else if (joy & JOYDOWN) {
		if (joy & JOYLEFT)
			joyBmp=38;
		else if (joy & JOYRIGHT)
			joyBmp=36;
		else
			joyBmp=37;
	} else {
		if (joy & JOYLEFT)
			joyBmp=39;
		else if (joy & JOYRIGHT)
			joyBmp=35;
		else
			joyBmp=32;
	}
	if (joyPtr[joyBmp-32])
		WinDrawBitmap(joyPtr[joyBmp-32], 136,136);
}

  /***************************************************/
 /* The actual event handlers (and Init/Exit)       */
/***************************************************/

/*
  C64 keyboard matrix:

    Bit 7   6   5   4   3   2   1   0
  0    CUD  F5  F3  F1  F7 CLR RET DEL
  1    SHL  E   S   Z   4   A   W   3
  2     X   T   F   C   6   D   R   5
  3     V   U   H   B   8   G   Y   7
  4     N   O   K   M   0   J   I   9
  5     ,   @   :   .   -   L   P   +
  6     /   ^   =  SHR HOM  ;   *   
  7    R/S  Q   C= SPC  2  CTL  <-  1
*/

/*
  Palm key codes are simply the position on the keyboard
0-16:   <- 1  2  3  4  5  6  7  8  9  0  +  -  P  HOM DEL F1
17-32: CTR Q  W  E  R  T  Y  U  I  O  P  @  *  ^  RST F3
33-48: R/S SL A  S  D  F  G  H  J  K  L  :  ;  =  RET F5
49-64: C= SHL Z  X  C  V  B  N  M  ,  .  / SHRCUD CLR F7 
65:             SPACE
*/

/*
  Tables for key translation
  Bit 0..2: row/column in C64 keyboard matrix
  Bit 3   : implicit shift
  Bit 5   : joystick emulation (bit 0..4: mask)
*/

const int key_byte[66] = {
	7, 7,  7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0,    0,
	   7,  7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, -1,    0,
	   7, -1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0,    0,
	       7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0, 0, 0,
	                      7
};

const int key_bit[66] = {
	1, 0,  3, 0, 3, 0, 3, 0, 3, 0, 3, 0, 3, 0, 3,  0,    4,
	   2,  6, 1, 6, 1, 6, 1, 6, 1, 6, 1, 6, 1, 6, -1,    5,
	   7, -1, 2, 5, 2, 5, 2, 5, 2, 5, 2, 5, 2, 5,  1,    6,
	       5, 7, 4, 7, 4, 7, 4, 7, 4, 7, 4, 7, 4,  7, 2, 3,
	                      4,
};

// ASCII to C=64 key number conversion table for ASCII values 0-127
// +128 indicates shifted key
UInt8 C64KeyNumberFromAscii[128] ={
//  0:     NUL     SOH     STX     ETX     EOT     ENQ     ACK     BEL
	   255,    255,    255,    255,    255,    255,    255,    255,
//  8:     BS      TAB     LF      VT      FF      CR      SO      SI
	    15,    255,     47,    255,    255,     47,    255,    255,
// 16:     DLE     DC1     DC2     DC3     DC4     NAK     SYN     ETB
	   255,    255,    255,    255,    255,    255,    255,    255,
// 24:     CAN     EM      SUB     ESC     FS,<-   GS,->   RS,^    US,v
	   255,    255,    255,    255,  128+63,    63, 128+62,    62,
// 32:     ' '      !       "       #       $       %       &       '
	    65,  1+128,  2+128,  3+128,  4+128,  5+128,  6+128,  7+128,
// 40:      (       )       *       +       ,       -       .       /
	 8+128,  9+128,     29,     11,     58,     12,     59,     60,
// 48:      0       1       2       3       4       5       6       7
	    10,      1,      2,      3,      4,      5,      6,      7,
// 56:      8       9       :       ;       <       =       >       ?
	     8,      9,     44,     45, 58+128,     46, 59+128, 60+128,
// 64:      @       A       B       C       D       E       F       G
	   255,     35,     55,     53,     37,     20,     38,     39,
// 72:      H       I       J       K       L       M       N       O
	    40,     25,     41,     42,     43,     57,     56,     26,
// 80:      P       Q       R       S       T       U       V       W
	    27,     18,     21,     36,     22,     24,     54,     19,
// 88:      X       Y       Z       [       \       ]       ^       _
	    52,     23,     51, 44+128,    255, 45+128,    255,    255,
// 64:             a       b       c       d       e       f       g
	   255,     35,    55,     53,     37,     20,     38,     39,
// 64:      h       i       j       k       l       m       n       o
	    40,     25,     41,     42,     43,     57,     56,     26,
// 64:      p       q       r       s       t       u       v       w
	    27,     18,     21,     36,     22,     24,     54,     19,
// 64:      x       y       z       {       |       }       ~      DEL
	    52,     23,     51,    255,    255,    255,    255,    15
};
// This function selects wether capital letters should be sent to the C=64 as shifted.
// Usefull for typing C=64 basic commands lO lI, but maybe confusing.
SetMixedCase(Boolean b)
{
	UInt8 i;
	
	for (i='A'; i<='Z'; i++)
		if (b)
			C64KeyNumberFromAscii[i] |= 128;
		else
			C64KeyNumberFromAscii[i] &= 0x7f;
}
// Pause/resume
SetPaused(Boolean b)
{
	if (b)
		StopSound();
	else if (sndStream)
		StartSound();
	pause=b;
}

/* Main event handler, handle on screen keyboard, menu. */
UInt16 inKey=0;
Boolean doFileForm=false, doQuicksave=false;
static Boolean HandleEventMain(EventPtr e)
{
	Boolean handled=false;
	UInt16 ctl;
	FormType *f=FrmGetActiveForm();
	RectangleType r;
	UInt16 i;
	UInt8 key;
	UInt32 j;

	/*if (e->eType>3) {
		StrPrintF(msg, "Got event %d", e->eType);
		SysFatalAlert(msg);
	}
	*/
	switch (e->eType) {
	  case menuEvent:
	    handled=true;
	    switch (e->data.menu.itemID) {
	      case GameFilemanager:
		// Give the armlet a chance to perform a snapshot
		// before displaying form.
	        doFileForm=true;
		armletData->cmd |= CMD68K_SAVE;
		//MenuEraseStatus(0);
		//FrmPopupForm(FileForm);
		break;
	      case GameQuicksave:
		doQuicksave=true;
		armletData->cmd |= CMD68K_SAVE;
		break;
	      case GameReset:
	        //SysFatalAlert("Reset");
		armletData->cmd |= CMD68K_RESET;
		break;
	      case GamePause:
	        SetPaused(!pause);
		break;
	      case OptionsAboutFrodo:
		ShowDialog(AboutForm);
		break;
	      case OptionsGraphicsandsound:
		ShowDialog(GfxSoundForm);
		break;
	      case OptionsJoystick:
		ShowDialog(JoystickForm);
		break;
	      case OptionsObscure:
		ShowDialog(ObscureForm);
		break;
	      case OptionsSwapjoysticks:
	        j=prefs->Joystick1On;
		prefs->Joystick1On=prefs->Joystick2On;
		prefs->Joystick2On=j;
		break;
	      case OptionsTogglemixedcase:
	        SetMixedCase(!mixedCase);
		break;
	      default:
		handled=false;
	    }
	    break;
	  case frmOpenEvent:
	    FrmDrawForm(f);
	    handled=true;
	    break;
	  case penDownEvent:
	    /* We need som custom event handling for the onscreen keyboard.
	       If we let the ctlEnterEvent past, FrmHandleEvent will go into a loop tracking
	       the pen until it's released, and the emulation will freeze.
	    */
	    for (i=MainFirstKeyGraphicButton; i<=MainSpaceGraphicButton; i++) {
	      FrmGetObjectBounds(f, FrmGetObjectIndex(f, i), &r);
	      if (RctPtInRectangle(e->screenX, e->screenY, &r))
		break;
	    }
	    // Keys are numbered right-to-left, top-to-bottom so space is last
	    if (i>MainSpaceGraphicButton)
	      break;
	    
	    inKey=i;
	    // TODO: Invert key
	    //if (inKey!=MainCommodoreGraphicButton && inKey!=MainShiftLockGraphicButton) {
	    if (inKey!=MainCommodoreGraphicButton && inKey!=MainShiftLeftGraphicButton) {
	        // On screen keyboard key down
	      armletData->key=swap32((KEYDOWN<<16) |
		(key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
		key_bit[inKey-MainFirstKeyGraphicButton]);
	    }
	    handled=true;
	    break;
	  case penUpEvent:
	    if (inKey) {
	      switch (inKey) {
		case MainCommodoreGraphicButton:
		  // Sticky commodore and left shift
		  cmdLock=!cmdLock;
	          armletData->key=swap32(((cmdLock? KEYDOWN:KEYUP)<<16) |
			(key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
			key_bit[inKey-MainFirstKeyGraphicButton]);
		  if (shiftLock && cmdLock)
		    lowerCase=!lowerCase;
		    
		  UpdateOnScreenKeyboard(f);
		  break;
		// case MainShiftLockGraphicButton:
		case MainShiftLeftGraphicButton:
		  shiftLock=!shiftLock;
		  armletData->key=swap32(((shiftLock? KEYDOWN:KEYUP)<<16) |
			(key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
			key_bit[inKey-MainFirstKeyGraphicButton]);
		  if (shiftLock && cmdLock)
		    lowerCase=!lowerCase;
		  
		  UpdateOnScreenKeyboard(f);
		  break;
		default:
		  // On-screen keyboard key up
		  armletData->key=swap32((KEYUP<<16) |
			(key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
			key_bit[inKey-MainFirstKeyGraphicButton]);
		  if (inKey==MainShiftRightGraphicButton && cmdLock) {
		    lowerCase=!lowerCase;
		    UpdateOnScreenKeyboard(f);
		  } else {
		    // TODO: Invert key
		  }
	      }
	      inKey=0;
	      handled=true;
	    }
	    break;
	  case keyDownEvent:
	    handled = false;
	    
	    /* We only care about some ascii characters.*/
	    if(e->data.keyDown.chr <0 || e->data.keyDown.chr > 127 )
	         break;

	    key = C64KeyNumberFromAscii[e->data.keyDown.chr];

	    /* character does not exist on the C64 keyboard/*/
	    if( key == 255)
		break;
	    
	    armletData->key = swap32((KEYDOWN<<16) |
	          ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) |
	            key_bit[key&0x7f]);

	    /* Need to wait some frames / some ticks for the KEYDOWN to be noticed.
	    Notice the time here.*/
	    ticksLastKeyDown = TimGetTicks();

	    /* savedGraffitiKey will be used later on */
	    savedGraffitiKey = swap32((KEYUP<<16) |
	          ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) |
	            key_bit[key&0x7f]);
		    
	    handled = true;
	    break;
	    // Try to pause when menu goes down
	  case winExitEvent:
	    if (e->data.winExit.exitWindow==(WinHandle) FrmGetFormPtr(MainForm)) {
		SetPaused(true);
	    }
	    break;
	  case winEnterEvent:
	    if (e->data.winEnter.enterWindow==(WinHandle) FrmGetFormPtr(MainForm) &&
	        e->data.winEnter.enterWindow==(WinHandle) FrmGetFirstForm()) {
		SetPaused(false);
	    }
	    break;
	}
	return handled;
}

// Joystick dialog event handler. Key sampling.

// Clear text for hardware keys
char *keys[33]={
	"Power","PgUp",	"PgDn",	"Key1",	"Key2",	"Key3",	"Key4",	"Cradle",
	"Antenna","Contrast","Unknown1","Unknown2","Unknown3","Unknown4","Unknown5","Unknown6",
	"Unknown7","Unknown8","Unknown9","Unknown10","Unknown11","Unknown12","Unknown13","Unknown14",
	"Unknown15","Unknown16","Unknown17","Unknown18","Unknown19","Unknown20","Unknown21","Unknown22",
	"Unset"
};

char *StringForKey(UInt32 k)
{
	int i;
	
	for (i=0; i<32; i++)
		if (k&(1<<i)) break;
	
	return keys[i];
}

// Klunky like hell... A requester and three buttons to get a key? Well, it was the easiest way.
static Boolean HandleEventSampleKey(EventPtr e)
{
	Boolean handled=false;
	FormType *f=FrmGetActiveForm();
	
	switch (e->eType) {
	  case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
	      case SampleKeySampleButton:
		sampledKey=KeyCurrentState();
		SetText(f, SampleKeyKeyLabel, StringForKey(sampledKey));
		//StrPrintF(msg, "Sampled: %s", StringForKey(sampledKey));
		//SysFatalAlert(msg);
		handled=true;
		break;
	    }
	    break;
	}
	return handled;
}

UInt32 dlgUp, dlgLeft, dlgRight, dlgDown, dlgFire;
void InitJoystickDialog(FormType *f)
{
	SetText(f, JoystickUpLabel, StringForKey(swap32(prefs->KeyUp)));
	SetText(f, JoystickLeftLabel, StringForKey(swap32(prefs->KeyLeft)));
	SetText(f, JoystickRightLabel, StringForKey(swap32(prefs->KeyRight)));
	SetText(f, JoystickDownLabel, StringForKey(swap32(prefs->KeyDown)));
	SetText(f, JoystickFireLabel, StringForKey(swap32(prefs->KeyFire)));
	SetValue(f, JoystickEmulate1Checkbox, prefs->Joystick1On? 1:0);
	SetValue(f, JoystickEmulate2Checkbox, prefs->Joystick2On? 1:0);
	SetValue(f, JoystickJoyPenScreenCheckbox,
		(swap32(prefs->JoyPen) & JOYPEN_SCREEN)? 1:0);
	SetValue(f, JoystickJoyPenGraffitiCheckbox,
		(swap32(prefs->JoyPen) & JOYPEN_GRAFFITI)? 1:0);
	dlgUp=swap32(prefs->KeyUp);
	dlgLeft=swap32(prefs->KeyLeft);
	dlgRight=swap32(prefs->KeyRight);
	dlgDown=swap32(prefs->KeyDown);
	dlgFire=swap32(prefs->KeyFire);
}
static Boolean HandleEventJoystick(EventPtr e)
{
	Boolean handled=false;
	FormType *f=FrmGetActiveForm();

	switch (e->eType) {
	  case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
	      case JoystickUpButton:
		dlgUp=SampleKey(dlgUp);
		SetText(f, JoystickUpLabel, StringForKey(dlgUp));
		handled=true;
		break;
	      case JoystickLeftButton:
		dlgLeft=SampleKey(dlgLeft);
		SetText(f, JoystickLeftLabel, StringForKey(dlgLeft));
		handled=true;
		break;
	      case JoystickRightButton:
		dlgRight=SampleKey(dlgRight);
		SetText(f, JoystickRightLabel, StringForKey(dlgRight));
		handled=true;
		break;
	      case JoystickDownButton:
		dlgDown=SampleKey(dlgDown);
		SetText(f, JoystickDownLabel, StringForKey(dlgDown));
		handled=true;
		break;
	      case JoystickFireButton:
		dlgFire=SampleKey(dlgFire);
		SetText(f, JoystickFireLabel, StringForKey(dlgFire));
		handled=true;
		break;
	      case JoystickOKButton:
		prefs->KeyUp=swap32(dlgUp);
		prefs->KeyLeft=swap32(dlgLeft);
		prefs->KeyRight=swap32(dlgRight);
		prefs->KeyDown=swap32(dlgDown);
		prefs->KeyFire=swap32(dlgFire);
		prefs->Joystick1On=GetValue(f, JoystickEmulate1Checkbox);
		prefs->Joystick2On=GetValue(f, JoystickEmulate2Checkbox);
		prefs->JoyPen=swap32(
		  (GetValue(f, JoystickJoyPenScreenCheckbox)? JOYPEN_SCREEN:0) |
		  (GetValue(f, JoystickJoyPenGraffitiCheckbox)? JOYPEN_GRAFFITI:0));
		SavePrefs();
		SetKeyMask();
		armletData->cmd |= CMD68K_SETPREFS;
		break;
	    }
	}
	return handled;
}

// Graphics and sound options
#define NSKIPLIST 6
char *frameskipList[NSKIPLIST] = {
	"Adaptive",
	"50 Hz",
	"25 Hz",
	"17 Hz",
	"13 Hz",
	"10 Hz"
};
#define NSCREENMODE 4
char *screenmodeList[NSCREENMODE] = {
	"Direct 16 (fastest)",
	"Direct 16 omap",
	"Direct 256",
	"Buffered 256 (most compatible)"
};
void InitGfxSoundDialog(FormPtr f)
{
	UInt16 idx;
	ListType *l;
	UInt32 fs,sm;
	
	idx=FrmGetObjectIndex(f, GfxSoundFrameskipList);
	if (idx!=frmInvalidObjectId) {
		l=(ListType *) FrmGetObjectPtr(f, idx);

		LstSetListChoices(l, frameskipList, NSKIPLIST);
		if (prefs->AdaptiveFrameSkip)
			fs=0;
		else
			fs=swap32(prefs->SkipFrames);
		if (fs>NSKIPLIST-1) fs=NSKIPLIST-1;
		LstSetSelection(l, fs);
		FrmCopyLabel(f, GfxSoundFrameskipPopTrigger, frameskipList[fs]);
	}
	idx=FrmGetObjectIndex(f, GfxSoundScreenModeList);
	if (idx!=frmInvalidObjectId) {
		l=(ListType *) FrmGetObjectPtr(f, idx);
		sm=swap32(prefs->ScreenMode);
		LstSetListChoices(l, screenmodeList, NSCREENMODE);
		if (sm<0 || sm>=NSCREENMODE) {
			StrPrintF(msg, "Bad screen mode %ld", sm);
			SysFatalAlert(msg);
			sm=0;
		}
		LstSetSelection(l, sm);
		FrmCopyLabel(f, GfxSoundScreenModePopTrigger, screenmodeList[sm]);
	}
	SetValue(f, GfxSoundSIDFiltersCheckbox, prefs->SIDFilters? 1:0);
	SetValue(f, GfxSoundTurboCheckbox, prefs->LimitSpeed? 0:1);
	SetValue(f, GfxSoundVolumeFeedbackSliderControl, swap32(prefs->Volume));
	if (!prefs->Volume)
		StrPrintF(msg, "Vol: Mute");
	else
		StrPrintF(msg, "Vol: %ld%%", (100*swap32(prefs->Volume))>>10);
	SetText(f, GfxSoundVolumeLabel, msg);
}
static Boolean HandleEventGfxSound(EventPtr e)
{
	UInt32 v;
	UInt16 idx;
	ControlPtr c;
	FormPtr f=FrmGetActiveForm();
	
	switch (e->eType) {
	  case ctlRepeatEvent:
	    v=e->data.ctlRepeat.value;
	    if (!v)
		StrPrintF(msg, "Vol: Mute");
	    else
		StrPrintF(msg, "Vol: %ld%%", (100*v)>>10);
	    SetText(f, GfxSoundVolumeLabel, msg);
	    /*
	    idx=FrmGetObjectIndex(f, GfxSoundVolumeLabel);
	    if (idx!=frmInvalidObjectId)
		if (c=FrmGetObjectPtr(f, idx))
			CtlDrawControl(c);
	    */
	    soundStruct.makeNoise=true;
	    SndStreamSetVolume(sndStream, v);
	}
	return false;
}
void ExitGfxSoundDialog(FormPtr f, uint16 choice)
{
	UInt16 idx;
	ListType *l;
	UInt32 v, sm;
	Boolean buffer=false;

	soundStruct.makeNoise=false;
	if (choice != GfxSoundOKButton) return;

	prefs->SIDFilters=GetValue(f, GfxSoundSIDFiltersCheckbox);
	prefs->LimitSpeed=!GetValue(f, GfxSoundTurboCheckbox);
	v=GetValue(f, GfxSoundVolumeFeedbackSliderControl);
	prefs->Volume=swap32(v);
	
	idx=FrmGetObjectIndex(f, GfxSoundFrameskipList);
	if (idx!=frmInvalidObjectId) {
		l=(ListType *) FrmGetObjectPtr(f, idx);
		idx=LstGetSelection(l);
		if (idx==0)
			prefs->AdaptiveFrameSkip=1;
		else if (idx>0 && idx<=NSKIPLIST-1) {
			prefs->AdaptiveFrameSkip=0;
			prefs->SkipFrames=swap32(idx);
		}
	}
	idx=FrmGetObjectIndex(f, GfxSoundScreenModeList);
	if (idx!=frmInvalidObjectId) {
		l=(ListType *) FrmGetObjectPtr(f, idx);
		idx=LstGetSelection(l);
		ChangeScreenMode(idx, false, false);
		if (FrmGetActiveForm()) FrmDrawForm(FrmGetActiveForm());
	}
	/*
	if (prefs->Volume && !canSound) {
		warning
	}
	*/
	if (v) {
		canSound=true;
		StartSound();
	} else
		StopSound();
		
	SavePrefs();
	armletData->cmd |= CMD68K_SETPREFS;
}

// "Obscure options" dialog. Cycles/frame, sprite collisions...
void InitObscureDialog(FormType *f)
{
	SetText(f, ObscureNormalCPLField, itoa(swap32(prefs->NormalCycles)));
	SetText(f, ObscureBadCPLField, itoa(swap32(prefs->BadLineCycles)));
	SetText(f, ObscureCIACPLField, itoa(swap32(prefs->CIACycles)));
	SetText(f, Obscure1541CPLField, itoa(swap32(prefs->FloppyCycles)));
	SetValue(f, ObscureSpritesCheckbox, prefs->SpritesOn? 1:0);
	SetValue(f, ObscureSpriteCollisionsCheckbox, prefs->SpriteCollisions? 1:0);
	SetValue(f, ObscureFastResetCheckbox, prefs->FastReset? 1:0);
	SetValue(f, ObscureCIAHackCheckbox, prefs->CIAIRQHack? 1:0);
	SetValue(f, ObscureEmulate1541Checkbox, prefs->Emul1541Proc? 1:0);
	// add REU size?
}
void ExitObscureDialog(FormType *f, UInt16 choice)
{
	Int32 i;
	UInt16 idx;
	ListType *l;
	
	if (choice != ObscureOKButton) return;
	
	i=atoi(GetText(f, ObscureNormalCPLField));
	prefs->NormalCycles=swap32(i);
	i=atoi(GetText(f, ObscureBadCPLField));
	prefs->BadLineCycles=swap32(i);
	i=atoi(GetText(f, ObscureCIACPLField));
	prefs->CIACycles=swap32(i);
	i=atoi(GetText(f, Obscure1541CPLField));
	prefs->FloppyCycles=swap32(i);
	prefs->SpritesOn=GetValue(f, ObscureSpritesCheckbox);
	prefs->SpriteCollisions=GetValue(f, ObscureSpriteCollisionsCheckbox);
	prefs->FastReset=GetValue(f, ObscureFastResetCheckbox);
	prefs->CIAIRQHack=GetValue(f, ObscureCIAHackCheckbox);
	prefs->Emul1541Proc=GetValue(f, ObscureEmulate1541Checkbox);
	
	SavePrefs();
	armletData->cmd |= CMD68K_SETPREFS;
}

// File manager. Has its own menu bar for selecting active drive unit.

// New version of ScanFiles. Use dynamic memory management, don't keep files open, cache.

UInt16 nfile=0, maxfile=0;
MemHandle fileH=NULL;
struct t_file {
	UInt16 card;   // 0xffff if internal pdb
	char name[32]; // file name in /Palm/programs/Frodo/
} *file=NULL;

UInt16 nitem=0, maxitem=0;
MemHandle itemH=NULL;
struct t_item {
	struct t_file *file;
	UInt16 index; // 0xffff if not a pdb
	char name[34];
} *item=NULL;

UInt16 nitemList=0, maxitemList=0;
MemHandle itemListH=NULL;
char **itemList=NULL;

FreeItems()
{
	int i;

	if (file) {
		if (fileH) MemHandleUnlock(fileH);
		file=NULL;
	}
	if (fileH) {
		MemHandleFree(fileH);
		fileH=NULL;
	}
	nfile=maxfile=0;

	if (item) {
		if (itemH) MemHandleUnlock(itemH);
		item=NULL;
	}
	if (itemH) {
		MemHandleFree(itemH);
		itemH=NULL;
	}
	nitem=maxitem=0;
	
	if (itemList) {
		if (itemListH) MemHandleUnlock(itemListH);
		itemList=NULL;
	}
	if (itemListH) {
		MemHandleFree(itemListH);
		itemListH=NULL;
	}
	nitemList=maxitemList=0;
}

Err Grow(MemHandle *h, MemPtr *p, UInt16 size, UInt16 n, UInt16 *max)
{
	Err e;

	if (!*h) {
		*max=16;
		*h=MemHandleNew(16L*size);
		if (!(*h)) return memErrNotEnoughSpace;
	} else if (n>*max) {
		*max*=2;
		if (*p) {
			MemHandleUnlock(*h);
			*p=NULL;
		}
		e=MemHandleResize(*h, ((UInt32)*max)*size);
		if (e) return e;
	} else {
		// No change
		return 0;
	}
	*p=MemHandleLock(*h);
	if (!*p) {
		MemHandleFree(*h);
		*h=NULL;
		return memErrNotEnoughSpace;
	}
	return 0;
}
Err AddFile(UInt32 card, char *fnam)
{
	Err err;

	err=Grow(&fileH, (MemPtr *)&file, sizeof(struct t_file), nfile+1, &maxfile);
	if (err) return err;

	file[nfile].card=card;
	MemSet(file[nfile].name, sizeof(file[nfile].name), 0x0);
	if (fnam) StrNCopy(file[nfile].name, fnam, sizeof(file[nfile].name)-1);

	nfile++;

	return 0;
}
Err AddItem(UInt16 idx, char *fnam)
{
	Err err;

	//SysFatalAlert("2");
	err=Grow(&itemH, (MemPtr *)&item, sizeof(struct t_item), nitem+1, &maxitem);
	if (err) return err;

	//SysFatalAlert("3");
	item[nitem].file=&(file[nfile-1]);
	item[nitem].index=idx;
	//SysFatalAlert("4");
	MemSet(item[nitem].name, sizeof(item[nitem].name), 0x0);
	if (fnam) {
		if (file[nfile-1].card==0xffff) {
			StrNCopy(item[nitem].name, fnam,
				sizeof(item[nitem].name)-3);
		} else {
			StrCopy(item[nitem].name, "X ");
			StrNCopy(item[nitem].name+2, fnam,
				sizeof(item[nitem].name)-3);
		}
		//SysFatalAlert("5");
	}
	nitem++;

	err=Grow(&itemListH, (MemPtr *) &itemList, sizeof(char *),
		nitemList+1, &maxitemList);
	//SysFatalAlert("6");
	if (err) return err;
	nitemList++;
	// No use setting the pointers yet, they will change
	//itemList[nitemList++]=item[nitem-1].name;
	
	//SysFatalAlert("7");

	return 0;
}
ScanFiles(FormPtr f)
{
	UInt16 i,j, n;
	MemHandle h;
	MemPtr p;
	ListPtr l;
	char *name=NULL, path[256], fnam[100], fnam2[100];
	UInt16 part=0, vol;
	UInt32 volIter, fIter;
	Err err;
	FileRef dir, file;
	FileInfoType info;
	DbFile dbf;
	
	// Just in case
	FreeItems();
	
	DbFileFromDbRef(mainDBRef, &dbf);
	err=Scan(&dbf, 0xffff, NULL);
	if (err) {
		StrPrintF(msg,"Out of memory when scanning files! (0,0x%x)", err);
		SysFatalAlert(msg);
		return;
	}
	// For each expansion card
	volIter=vfsIteratorStart;
	while (volIter!=vfsIteratorStop) {
		err=VFSVolumeEnumerate(&vol, &volIter);
		if (err) continue;
	
		//StrPrintF(msg, "Found expansion card %x", vol);
		//SysFatalAlert(msg);
		err=VFSFileOpen(vol, frodoDir, vfsModeRead, &dir);
		if (err) {
			//StrPrintF(msg, "Error opening Frodo dir: %x", err);
			//SysFatalAlert(msg);
			continue;
		}

		info.nameP=fnam;
		info.nameBufLen=sizeof(fnam);
		//SysFatalAlert("Card has /Palm/programs/Frodo");

		// For each file in /Palm/Programs/Frodo
		fIter=expIteratorStart;
		while (fIter!=expIteratorStop) {
			err=VFSDirEntryEnumerate(dir, &fIter, &info);
			if (err) continue;
			
			//StrPrintF(msg, "Found \"%s\"", fnam);
			//SysFatalAlert(msg);
			if (StrLen(fnam)<4) continue;
			if (!StrCompare(fnam+StrLen(fnam)-4, ".pdb")) {
				// Try to find cached index
				StrCopy(fnam2, fnam);
				StrCopy(fnam2+StrLen(fnam2)-4, ".idx");
				StrPrintF(path, "%s/%s", frodoDir, fnam2);
				err=VFSFileOpen(vol, path, vfsModeRead, &file);
				if (!err)  {
					//SysFatalAlert("Index found");
					err=ScanIndex(file, vol, fnam);
					if (err) {
						StrPrintF(msg,"Out of memory when scanning files! (1,0x%x)", err);
						SysFatalAlert(msg);
						break;
					}
					continue;
				}
				//SysFatalAlert("No index");
				StrPrintF(path, "%s/%s", frodoDir, fnam);
				err=VFSFileOpen(vol, path, vfsModeRead, &file);
				if (err) continue;

				//SysFatalAlert("Scanning...");
				DbFileFromFileRef(file, &dbf);
				err=Scan(&dbf, vol, fnam);
				VFSFileClose(file);
				if (err) {
					StrPrintF(msg,"Out of memory when scanning files! (2,0x%x)", err);
					SysFatalAlert(msg);
					break;
				}
			} else if (!StrCaselessCompare(fnam+StrLen(fnam)-4, ".d64") ||
			           !StrCaselessCompare(fnam+StrLen(fnam)-4, ".t64")) {
				//SysFatalAlert("d64/t64");
				// Add this one file
				err=AddFile(vol, fnam);
				if (err) {
					StrPrintF(msg,"Out of memory when scanning files! (3,0x%x)", err);
					SysFatalAlert(msg);
					break;
				}
				err=AddItem(0xffff, fnam);
				if (err) {
					StrPrintF(msg,"Out of memory when scanning files! (4,0x%x)", err);
					SysFatalAlert(msg);
					break;
				}
			}
		}
		VFSFileClose(dir);
	}
	//StrPrintF(msg, "Total %d items", nitemList);
	//SysFatalAlert(msg);
	if (itemList) {
	    if (nitem!=nitemList) {
		StrPrintF(msg, "%d!=%d !!!", nitem, nitemList);
		SysFatalAlert(msg);
	    }
	    for (i=0; i<nitem && i<nitemList; i++)
		itemList[i]=item[i].name;
	    i=FrmGetObjectIndex(f, FileFileList);
	    if (i!=frmInvalidObjectId) {
		l=(ListType *) FrmGetObjectPtr(f, i);

		LstSetListChoices(l, itemList, nitemList);
		LstDrawList(l);
	    }
	}
	//SysFatalAlert("Done");
}

// VFSFileRead as it ought to work :-)
Err VFSRead(FileRef f, UInt32 n, void *vP)
{
	char *buf=(char *)vP;
	UInt32 r, tot=0;
	Err err;

	while (tot<n) {
		err=VFSFileRead(f, n-tot, buf+tot, &r);
		if (err) return err;
		tot+=r;
	}
	return 0;
}

Err ScanIndex(FileRef f, UInt32 card, char *fnam)
{
	UInt16 idx;
	UInt32 n;
	char name[32];
	Err err;

	err=AddFile(card, fnam);
	if (err) return err;
	while (!VFSFileEOF(f)) {
		err=VFSRead(f, 2, &idx);
		if (err) return err;
		err=VFSRead(f, 32, name);
		if (err) return err;
		err=AddItem(idx, name);
	}
	return 0;
}
Err Scan(DbFile *dbf, UInt32 card, char *fnam)
{
	UInt16 i, n;
	UInt32 type, app;
	MemHandle h;
	MemPtr p;
	RectangleType r={{20,70},{120,20}};
	char name[32]="";
	FileRef index=NULL;
	Err err;
	char s[37];

	if (!dbf) return 0;

	DbFileInfo(dbf, &type, &app, &n);

/*
	if (app!=appSignature ||
	    (type!=mainType && type!=dbDataType)) {
		StrPrintF(msg, "Wrong app or type %c%c%c%c %c%c%c%c",
			(app>>24)&0xff, (app>>16)&0xff, (app>>8)&0xff, app&0xff,
			(type>>24)&0xff, (type>>16)&0xff, (type>>8)&0xff, type&0xff);
		SysFatalAlert(msg);
		return 0;
	}
*/
	err=AddFile(card, fnam);
	if (err) return err;
	//StrPrintF(msg, "nfile=%d, maxfile=%d, h=0x%lx, p=0x%lx, size=%ld",
	//	nfile, maxfile, fileH, file, MemHandleSize(fileH));
	//SysFatalAlert(msg);
	if (fnam && card!=0xffff) {
		StrPrintF(msg, "%s/%s", frodoDir, fnam);
		StrCopy(msg+StrLen(msg)-4, ".idx");
		err=VFSFileOpen(card, msg, vfsModeWrite | vfsModeCreate | vfsModeTruncate,
			&index);
		if (err) {
			index=0;
			StrPrintF(msg, "Failed to create index: %x", err);
			SysFatalAlert(msg);
		}
	}

	// Scan a file for programs
 
	//SysFatalAlert("File has data");
	for (i=0; i<n; i++) {
		err = DbFileGetRecord(dbf, i, &h);

		WinEraseRectangle(&r, 0);
		WinDrawRectangleFrame(simpleFrame, &r);
		StrPrintF(msg, "Db record %d/%d", i, n);
		WinDrawTruncChars(msg, StrLen(msg),
			r.topLeft.x+10, r.topLeft.y+5,r.extent.x-20);

		//StrPrintF(msg, "Record %d=0x%lx", i, h);
		//SysFatalAlert(msg);
		if (!h) continue;
		
		if (p=MemHandleLock(h)) {
			StrNCopy(s, p, 36);
			s[35]='\0'; 
			//StrPrintF(msg, "Record %d is \"%s\"", i, s);
			//SysFatalAlert(msg);
			if (!StrNCompare(p, "SNAP", 4) ||
		    	    !StrNCompare(p, "SPLT", 4)) {
				err=AddItem(i, s+4);
				//SysFatalAlert("1");
				if (err) {
					MemHandleUnlock(h);
					DbFileReleaseRecord(dbf, i, h, false);
					if (index) VFSFileClose(index);
					return err;
				}
				if (dbf->external) {
				    StrCopy(name,item[nitem-1].name+2);
				} else {
				    StrCopy(name,item[nitem-1].name);
				}
				if (index) {
					VFSFileWrite(index, 2, &i, NULL);
					VFSFileWrite(index, 32, name, NULL);
				}
			} else {
				if (!StrCompare(name, "")) {
					StrPrintF(msg, "Expected a SNAP "\
						"@ %d, not %s",
						i, s);
					SysFatalAlert(msg);
				} else if (StrNCompare(name, p+4, 32)) {
					StrPrintF(msg, "Expected a %s part"\
						" @ %d, not %s",
						name, i, s);
					SysFatalAlert(msg);
				}
			}
			MemHandleUnlock(h);
		} else  {
			//StrPrintF(msg, "Couldn't lock record %d", i);
			//SysFatalAlert(msg);
		}
		DbFileReleaseRecord(dbf, i, h, false);
	} // end for each record

	if (index) VFSFileClose(index);

	return 0;
}
static Boolean HandleEventFile(EventPtr e)
{
	Boolean handled=true, doExit=false;
	FormPtr f=FrmGetActiveForm();
	ListPtr l=NULL;
	UInt16 i;
	char *s;
	FileRef fr;
	DbFile dbf;
	Err err;
	
	switch (e->eType) {
	  case frmOpenEvent:
	    FrmDrawForm(f);
	    ScanFiles(f);
	    SetPaused(true);
	    break;
	  case ctlSelectEvent:
	    i=FrmGetObjectIndex(f, FileFileList);
	    if (i!=frmInvalidObjectId)
	      l=(ListType *) FrmGetObjectPtr(f, i);
            
	    switch (e->data.ctlSelect.controlID) {
	      case FileLoadButton:
	        if (l) {
			i=LstGetSelection(l);
			if (i>=0) {
				if (item[i].file->card==0xffff) {
					// Internal pdb
					DbFileFromDbRef(mainDBRef, &dbf);
					Load(&dbf, item[i].index);
				} else if (item[i].index==0xffff) {
					// external file
					//err=LoadFile(item[i].file->card, item[i].file->name);
				} else {
					// external pdb
					StrPrintF(msg, "%s/%s", frodoDir,
						item[i].file->name);
					err=VFSFileOpen(item[i].file->card, msg,
						vfsModeRead, &fr);
					if (!err) {
						DbFileFromFileRef(fr, &dbf);
						Load(&dbf, item[i].index);
					}
					VFSFileClose(fr);
				}
			}
		}
		doExit=true;
		break;
	      case FileSaveButton:
		if (s=ShowSaveAsDialog(NULL)) {
			// When the file dialog is about to go up, the C=64 is alerted
			// to save a snapshot, then we delay a frame.
			// Thanks to this armletData->snapshot is valid here.
			SaveSnapshot(s, -1);
		}
		ScanFiles(f);
		break;
	      case FileDeleteButton:
	        // Delete by index (removes all subsequent parts up to the next
		// valid "first part")
	        if (l) {
			i=LstGetSelection(l);
			if (i>=0 && item[i].file->card==0xffff) {
				DbFileFromDbRef(mainDBRef, &dbf);
				Delete(&dbf, item[i].index);
				ScanFiles(f);
			}
		}
		break;
	      case FileExitButton:
	        doExit=true;
		break;
	    }
	  default:
	    handled=false;
	}
	
	if (doExit) {
		FreeItems();
		SetPaused(false);
		FrmReturnToForm(MainForm);
	}
	return handled;
}

// Application event handler. Handle form init.

static Boolean AppHandleEvent(EventPtr e)
{
	Boolean handled=false;
	FormType *f=FrmGetActiveForm();
	UInt16 fid;

	/*if (e->eType>3) {
		StrPrintF(msg, "Got event %d", e->eType);
		SysFatalAlert(msg);
	}*/
	switch (e->eType) {
	  case frmLoadEvent:
	    switch (e->data.frmLoad.formID) {
	      case MainForm:
		f=FrmInitForm(e->data.frmLoad.formID);
		FrmSetActiveForm(f);
		FrmSetEventHandler(f, HandleEventMain);
		handled=true;
		break;
	      case FileForm:
		f=FrmInitForm(FileForm);
		FrmSetActiveForm(f);
		FrmSetEventHandler(f, HandleEventFile);
		handled=true;
		break;
	    }
	    break;
	  case appStopEvent:
	    FrmEraseForm(f);
	    FrmDeleteForm(f);
	    exiting = true;
	    handled = true;
	    break;
	// Check pen position, I don't trust non-pen events.
	  case penDownEvent:
	  case penMoveEvent:
	    penDown=true;
	    penX=e->screenX;
	    penY=e->screenY;
	    break;
	  case penUpEvent:
	    penDown=false;
	    break;
	}
	return handled;
}

  /***********************************************/
 /* ARM -> 68k callback functions, esp. VBlank. */
/***********************************************/

// Called every frame by the armlet. Our only chance to handle I/O.
UInt32 VBlank68k(void *p)
{
	static UInt32 startTick=0, frame=0, cnt=0, tps, tpf, resets=0;
	EventType e;
	static UInt32 perfTick=0, didSkip=0, didDraw=0;
	UInt32 tick, sleepTo=0;
	Int32 diff;
	UInt32 fillAt, i;
	static UInt32 lastTick=0;
	static UInt8  nskipped=0;
	Boolean handled;
	UInt16 menuError;
	RectangleType r;
	char *s;

	// Initialization at frame==0
	if (frame==0L) {
		tps=SysTicksPerSecond();
		tpf=tps/50;
		// Timing mismatch if tps isn't a multiple of fps
		diff=100*tps - tpf*5000;
		if (diff>3 || diff<-3) { // >3% off optimal
			StrPrintF(msg, "Your SysTicksPerSecond() (%ld) will result in "\
				"very bad timing, be warned.", tps);
			SysFatalAlert(msg);
		}
		startTick=TimGetTicks();
	}
	// Show the about dialog first time, after the C=64 has initializeditself
	if (firstTime && frame==100)
		ShowDialog(AboutForm);

	// I do this first to get maximum precision in the clocks.
	// sleepTo is when this frame should end, in a perfect world.
	// It's lastTick+1/50 s. If we lag very badly (10 frames right now),
	// the timing is reset.

	if (!lastTick || nskipped>=10) {
		// First time, or after pause/dialog, or just a slow cpu.
		lastTick=tick=sleepTo=TimGetTicks();
		nskipped=0;
		soundStruct.underflow=0;
		resets++;
	} else {
		tick=TimGetTicks();
		sleepTo=lastTick+tpf;
	}
	
	// Send hint to armlet to skip next frame if we're running late
	armletData->cmd &= ~(CMD68K_DRAWFRAME);
	if (!prefs->LimitSpeed) {
		// Turbo mode, no sleep
		sleepTo=tick;
	}
	if (prefs->AdaptiveFrameSkip) {
		// Adaptive mode, draw next frame if we're on time, otherwise skip.
		if (tick <= sleepTo)
			armletData->cmd |= CMD68K_DRAWFRAME;
	} else {
		// Old-style, use user-specified frame skip ratios.
		if (nskipped+1 >= swap32(prefs->SkipFrames))
			armletData->cmd |= CMD68K_DRAWFRAME;
	}
	if (armletData->cmd & CMD68K_GETPREFS) {
		// We're unsure of the prefs (esp. screen mode), so don't draw anything
		// until we get a set of prefs.
		armletData->cmd &= ~CMD68K_DRAWFRAME;
	}
	if (armletData->cmd & CMD68K_DRAWFRAME) {
		nskipped=0;
		didDraw++;
	} else {
		nskipped++;
		didSkip++;
	}
	
	// Debug, show PC in GUI
	/*
	StrPrintF(msg, "PC: 0x%04lx", swap32(armletData->pc));
	WinDrawChars(msg, StrLen(msg), 100,140);
	*/
	// Debug, show SID in GUI
	/*
	if ((frame&63)==0) {
		r.topLeft.x=100; r.topLeft.y=136;
		r.extent.x=60; r.extent.y=24;
		WinEraseRectangle(&r, 0);
		StrPrintF(msg, "%ld", swap32(armletData->nsamp));
		WinDrawChars(msg, StrLen(msg), 100, 137);
		StrPrintF(msg, "%ld", swap32(armletData->max));
		WinDrawChars(msg, StrLen(msg), 100, 148);
		StrPrintF(msg, "%ld", swap32(armletData->min));
		WinDrawChars(msg, StrLen(msg), 130, 148);
		armletData->nsamp=0;
		armletData->max=swap32(-10000L);
		armletData->min=swap32(10000L);
		
	}*/
	/*
	for (i=0; i<65; i++) {
		WinDrawPixel(90+i, 148 + swap16(armletBuf[i<<4])>>12);
	}
	*/
	// Debuggish, performance
	if ((frame & 31)==0) {
		r.topLeft.x=100; r.topLeft.y=136;
		r.extent.x=36; r.extent.y=24;
		WinEraseRectangle(&r, 0);
		
		// Emulation speed
		// Should really display CPU speed when turbo on
		if (!prefs->LimitSpeed) {
			// Unlimited speed. Show actual emulation speed compared to C64
			StrPrintF(msg, "%ld%%", 100L*32*tpf/(tick-perfTick));
		} else if (prefs->AdaptiveFrameSkip) {
			// Adaptive frame skip. Show % of frames drawn.
			StrPrintF(msg, "%ld%%", 100*didDraw/32);
		} else {
			// Manual frame skip, no turbo. Show... What?
			StrPrintF(msg, "%ld%%", 100*didDraw*swap32(prefs->SkipFrames)/32);
		}
		WinDrawChars(msg, StrLen(msg), 100, 137);
		perfTick=tick;
		didSkip=didDraw=0;
	}
	
	frame++;
	
	// Display
	if ((armletData->cmd & CMD68K_DRAWFRAME) && bmpBits) {
		// Not rendering directly to screen, so blit bitmap
		// We need to handle density better here
		//WinDrawBitmap((BitmapPtr)bmpv3, (screenW-320)/2, 0);
		//for (i=0; i<16; i++)
		//	bmpBits[i]=i;
		WinDrawBitmap((BitmapPtr)bmpv3, (screenW-320)/4, 0);
	}
	// Shovel prefs
	if (armletData->cmd & CMD68K_GETPREFS) {
		// We ordered a set of prefs, well, here they are!
		armletData->cmd &= ~(CMD68K_GETPREFS);
		if (prefs->Volume && !canSound) {
			prefs->Volume=0;
			// In futurity, we should let prefs override Palmos, since apparently
			// sound detection in 5.0 DOES NOT WORK AT ALL on the Tungsten T 1.0
			// So a "Are you sure", then set canSound to true.
		}
		if (armletData->cmd & CMD68K_GETPREFS) SysFatalAlert("BAD MOJO!");
		// Try to get default screen mode based on device id
		prefs->ScreenMode=swap32(SCREENMODE_16);
		if (DeviceHasOMAPCPU()) prefs->ScreenMode=swap32(SCREENMODE_16ALT);
		ChangeScreenMode(swap32(prefs->ScreenMode), false, true);
		if (FrmGetActiveForm()) FrmDrawForm(FrmGetActiveForm());
		SetKeyMask();
		//DumpPrefs(prefs);
		SavePrefs();
	}
	// Snapshot creation (File manager and quicksave)
	if (armletData->cmd & CMD68K_SAVE) {
		// The file manager asked for a snapshot before displaying.
		// Nothing special needs to be done
		armletData->cmd &= ~CMD68K_SAVE;
	}
	
	// Sound, don't overqueue in turbo mode
	/*
	if (soundStruct.sampleOffset > SOUND_BUFSIZE) {
		StrPrintF(msg, "smplOfs: %ld > %ld", soundStruct.sampleOffset, SOUND_BUFSIZE);
		SysFatalAlert(msg);
	}
	if (soundStruct.samplesQueued > SOUND_BUFSIZE) {
		StrPrintF(msg, "smplQ: %ld > %ld", soundStruct.samplesQueued, SOUND_BUFSIZE);
		SysFatalAlert(msg);
	}
	*/
	if (canSound && prefs->Volume &&
	    soundStruct.samplesQueued+FRAME_SAMPLES<=SOUND_BUFSIZE) {
		// Explanation:
		//   soundBuf is a large(ish) circular buffer that's filled here
		//   with data from the armlet and drained by the sound callback.
		//   SampleOffset is the offset (in samples) to the next sample
		//   that will be played. SamplesQueued is the number of samples
		//   beyond this point that are ready for consumption.
		
		fillAt=soundStruct.sampleOffset+soundStruct.samplesQueued;
		if (fillAt > SOUND_BUFSIZE)
			fillAt -= SOUND_BUFSIZE;
			
		if (fillAt > SOUND_BUFSIZE) {
			StrPrintF(msg, "%ld > %ld", fillAt, SOUND_BUFSIZE);
			SysFatalAlert(msg);
		}
		if (fillAt+FRAME_SAMPLES > SOUND_BUFSIZE) {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				(SOUND_BUFSIZE-fillAt)*2);
			MemMove(soundStruct.soundBuf,
				armletBuf+SOUND_BUFSIZE-fillAt,
				(fillAt+FRAME_SAMPLES-SOUND_BUFSIZE)*2);
		} else {
			MemMove(soundStruct.soundBuf+fillAt, armletBuf,
				FRAME_SAMPLES*2);
		}
		soundStruct.samplesQueued += FRAME_SAMPLES;
		/*
		if (frame==100) {
			StrPrintF(msg, "Sample: %x", armletBuf[0]);
			SysFatalAlert(msg);
		}*/
	}
	if (canSound && prefs->Volume)
		soundStruct.underflow -= FRAME_SAMPLES;


	// The File form delays one frame, so that the armlet can make a snapshot first.
	if (doFileForm) {
		doFileForm=false;
		MenuEraseStatus(0);
		FrmPopupForm(FileForm);
	}
	// Same with quicksave
	if (doQuicksave) {
		int l=StrLen(lastName), sl, i, n=0,m=1;
		char serial[8];
		doQuicksave=false;
		MenuEraseStatus(0);
		
		// We should make up a default name, such as "Game quicksave 1"
		// Clippy style 'AI' follows
	
		// strip .d64/.t64 if present
		if (l>=4 && lastName[l-4]=='.') {
			l-=4;
		}
		// read serial if present
		for (i=l-1; i>=0; i--) {
			if (lastName[i]<'0' || lastName[i]>'9')
				break;
			n+=m*(lastName[i]-'0');
			m*=10;
		}
		l=i+1;
		if (n==0) n=1;
		// append serial+1 to name
		if (l<31 && (l==0 || lastName[l-1]!=' ')) // Insert space if appropriate
			StrPrintF(serial, " %d", n+1);
		else
			StrPrintF(serial, "%d", n+1);
		sl=StrLen(serial);
		if (l+sl>31) // Truncate lastLoad if nessecary
			l=31-sl;
		StrCopy(lastName+l, serial); // add serial
			
		s=ShowSaveAsDialog(lastName);
		if (s) {
			SaveSnapshot(s, -1);
		}
	}

	// No key events
	armletData->key=swap32(0x00ffff);

	// Handle queued up events
	do {
		UInt16 menuError;
		
		tick=TimGetTicks();
		// Mind signed/unsigned 16/32 issues!
		// This simple line caused me a lot of grief :-)
		if (sleepTo > tick+tps) {
			// Sleep more than a second? Something very strange just happened....
			// int32 roll-over? Second coming of Christ?
			sleepTo=tick;
			resets++;
		}
		// Get event, sleepTo-tick is the max we should wait
		if (pause)
			EvtGetEvent(&e, evtWaitForever);
		else if (tick >= sleepTo)
			EvtGetEvent(&e, 0);
		else
			EvtGetEvent(&e, sleepTo-tick);

		// return to armlet if no more events queued
		if (e.eType==nilEvent && !pause) break;

		cnt++;

		// Standard Palm Event handling
		if (!SysHandleEvent(&e))
			if (!MenuHandleEvent(0,&e,&menuError))
				if (!AppHandleEvent(&e))
					FrmDispatchEvent(&e);
	} while (e.eType!=appStopEvent);
	
	// Joystick
	armletData->joystick=0;

	CheckJoyPen();
	CheckJoyKey();
	
	UpdateOnScreenJoystick();
		
	// Graffiti key-up
	// Time to enqueue penUpEvent? Wait some ticks after KEYDOWN then --> KEYUP
	if(ticksLastKeyDown && ticksLastKeyDown+15 > tick &&
	   armletData->key==swap32(0x00ffff)) {
		armletData->key=savedGraffitiKey;
		ticksLastKeyDown=0;
	}

	
	
	// Normal exit, create auto snapshot
	if (e.eType==appStopEvent) {
		armletData->errCode=swap32(ERR_NORMAL_EXIT);
		armletData->cmd |= CMD68K_SAVE;
		FreeItems();
	}
	lastTick=sleepTo;
	return 0;
}

// The JoyPen joystick emulation uses the pen to specify joystick direction.
// Basically, you have an origin and and extent, then see which octant the
// pen is in.

UInt32 CheckJoyPen2(int32 x, int32 y, int32 sz)
{
	Int32 dx=penX-x-(sz>>1), dy=penY-y-(sz>>1);
	Int32 ax=dx, ay=dy;
	//static Boolean downBefore;

	// In the zone?
	if (!penDown || penX<x || penY<y || penX>=x+sz || penY>=y+sz) {
		//downBefore=false;
		return 0;
	}

	// Tap-fire
	// ...isn't quite as good as I thought it'd be. Set a hw button instead.
	// if (!downBefore) f=swap32(JOYFIRE); 
	//downBefore=true;
	
	// ax, ay = abs(dx), abs(dy)
	if (ax<0) ax=-ax;
	if (ay<0) ay=-ay;

	// If we're in the inner 16'th then call it centered joystick.
	// Not. Too confusing. Just lift the pen.
	/*center=sz>>4;
	if (center==0) center=2;
	if (ax<center && ay<center) return 0;
	*/
	// tan v=sin v/cos v = y/x
	// tan pi/8 = y/x = 0.4142
	// x*4142 = y*10000
	// Ie. in the below means "In the two left-right octants"
	if (ax*4142 > ay*10000) {
		// Left, right
		if (dx<0)
			return swap32(JOYLEFT);
		else
			return swap32(JOYRIGHT);
	} else if (ay*4142 > ax*10000) {
		// Up, down
		if (dy<0)
			return swap32(JOYUP);
		else
			return swap32(JOYDOWN);
	} else if (dx<0) {
		// UpLeft, DownLeft
		if (dy<0)
			return swap32(JOYUP | JOYLEFT);
		else
			return swap32(JOYDOWN | JOYLEFT);
	} else {
		// UpRight, DownRight
		if (dy<0)
			return swap32(JOYUP | JOYRIGHT);
		else
			return swap32(JOYDOWN | JOYRIGHT);
	}
}
// Joystick emulation using pen.
CheckJoyPen()
{
	if (prefs->JoyPen & swap32(JOYPEN_SCREEN)) {
		// Main screen
		armletData->joystick |= CheckJoyPen2(0,0,160);
	}
	if (prefs->JoyPen & swap32(JOYPEN_GRAFFITI)) {
		// Bottom right (on the joystick indicator (which isn't there yet))
		armletData->joystick |= CheckJoyPen2(136, 136, 24);
	}
}

// Joystick keys on the keyboard
CheckJoyKey()
{
	UInt32 u, l, r, d, f, k, j;

	j=swap32(armletData->joystick);
	u=swap32(prefs->KeyUp);
	l=swap32(prefs->KeyLeft);
	r=swap32(prefs->KeyRight);
	d=swap32(prefs->KeyDown);
	f=swap32(prefs->KeyFire);

	k=KeyCurrentState();

	// Directions may have been partially set by JoyPen
	if (k&u) j|=JOYUP;
	if (k&l) j|=JOYLEFT;
	if (k&r) j|=JOYRIGHT;
	if (k&d) j|=JOYDOWN;
	if (k&f) j|=JOYFIRE;

	armletData->joystick=swap32(j);
}

// Guess what this does.
SavePrefs()
{
	MemHandle h;
	MemPtr p;
	int i;
	
	if (!prefs) {
		SysFatalAlert("No prefs to save...");
		return 5;
	}
	// Check if somehow a nuked prefs made it here.
	if (IsNuked(prefs)) {
		SysFatalAlert("Prefs is nuked, won't save");
		return 6;
	}

	if (!(h=DmGetResource(prefsType, IDX_PREFS))) {
		if (!exiting) SysFatalAlert("Error saving preferences");
		return 1;
	}
	if (MemHandleSize(h) < sizeof(struct PrefsState))  {
		DmReleaseResource(h);
		if (!exiting) SysFatalAlert("Prefs resource to small?!");
		return 2;
	} 
	if (!(p=MemHandleLock(h))) {
		DmReleaseResource(h);
		if (!exiting) SysFatalAlert("Couldn't lock prefs!");
		return 3;
	} 
	// And finally write the prefs.
	if (DmWrite(p, 0, prefs, sizeof(struct PrefsState))
			!= errNone) {
		MemHandleUnlock(h);
		DmReleaseResource(h);
		if (!exiting) SysFatalAlert("Error writing preferences");
		return 4;
	}
	MemHandleUnlock(h);
	DmReleaseResource(h);
	return 0;
}

// VBlank stuff ends here.

// Display one of the requesters. Just ugly FatalAlerts right now.
Int32 ShowRequester68k(Int32 *p)
{
	Int32 *ip, id, arg1, arg2, arg3, arg4;
	char buf[11];
	int i;
	
	if (!p) {
		StrPrintF(msg, "Illegal call to ShowRequester, parameter is NULL");
		SysFatalAlert(msg);
		return 0;
	}
	ip=(Int32 *)swap32(p);
	id=swap32(ip[0]);
	arg1=swap32(ip[1]); arg2=swap32(ip[2]);
	arg3=swap32(ip[3]); arg4=swap32(ip[4]);
	
	switch (id) {
		case REQ_BAD_SNAPSHOT:
			StrPrintF(msg, "Not a valid Frodo snapshot (\"%s\").", arg1);
			break;
		case REQ_BAD_SNAPSHOT_VERSION:
			StrPrintF(msg, "Invalid snapshot version %ld.", arg1);
			break;
		case REQ_BAD_OPCODE:
			StrPrintF(msg, "Illegal opcode 0x%04x at 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP:
			StrPrintF(msg, "Jump to I/O space at 0x%04x to 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_OPCODE_1541:
			StrPrintF(msg, "1541: Illegal opcode 0x%04x at 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP_1541:
			StrPrintF(msg, "1541: Jump to I/O space at 0x%04x to 0x%04x.",
				((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_NEW_DRIVE:
			StrPrintF(msg, "Out of memory changing disk %d.",
				((UInt16)arg1));
			break;
		case REQ_NEW_SIDRENDERER:
			StrPrintF(msg, "Out of memory when starting audio.");
			break;
		// DEBUG requesters below this point
		case REQ_DEBUG:
			StrPrintF(msg, "Debug: %lx %lx %lx %lx",
				arg1, arg2, arg3, arg4);
			break;
		case REQ_DEBUGSTR:
			StrPrintF(msg, "Debug \"%s\" %lx %lx %lx",
				arg1, arg2, arg3, arg4);
			break;
		case REQ_EL_TEXT:
			for (i=0; i<10; i++)
				buf[i]=((char *)arg2)[i];
			buf[10]='\0';
			StrPrintF(msg, "EL_*_TEXT: %lx (screen=%lx) <- %lx (\"%s\"), shift=%ld",
				arg1, screenBits, arg2, buf, arg3);
			break;
		case REQ_EL_GFX:
			for (i=0; i<10; i++)
				buf[i]=((char *)arg2)[i];
			buf[10]='\0';
			StrPrintF(msg, "EL_*: %lx (screen=%lx) <- %lx, shift=%ld",
				arg1, screenBits, arg2, arg3);
			break;
		deafult:
			StrPrintF(msg, "Unknow req %ld (%ld,%ld,%ld,%ld)",
				arg1, arg2, arg3, arg4);
	}
	// TODO 1541 errors should have two buttons "Reset C64", "Reset 1541"
	// I really should draw real Alrt's for these...
	SysFatalAlert(msg);

	return 0;
}

// new, delete for armlet side c++ (check palmposix.c)

void *Malloc68k(UInt32 n)
{
	void *retval;
	n=swap32(n);
	retval=MemPtrNew(n);
	//StrPrintF(msg, "Alloc(%ld) -> %lx", n, retval);
	//SysFatalAlert(msg);
	return retval;
}

UInt32 Free68k(void *p)
{
	p=(void *)swap32(p);
	//StrPrintF(msg, "Free(%lx)", p);
	//SysFatalAlert(msg);
	MemPtrFree(p);
	return 0;
}
Boolean DeviceHasPXACPU()
{
	UInt32 deviceID=0;
	Boolean hasPXACPU=false;
	
	// Should never fail
	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);
	
	
	if(deviceID == sonyHwrOEMDeviceID_NX70 ||
	   deviceID == sonyHwrOEMDeviceID_NZ90 ||
	   deviceID == sonyHwrOEMDeviceID_TG50 ||
	   deviceID == sonyHwrOEMDeviceID_NX73 ||
	   deviceID == sonyHwrOEMDeviceID_NX80 ||
	   deviceID == palmDeviceID_TC ||
	   deviceID == palmDeviceID_T3)
	   hasPXACPU = true;
	   
	   return hasPXACPU;
}

Boolean DeviceHasOMAPCPU()
{
	UInt32 deviceID=0;
	Boolean hasOMAPCPU=false;
	
	// Should never fail
	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);
	
	
	if(deviceID == palmDeviceID_TT	 	||
	   deviceID == palmDeviceID_T2	 	||
	   deviceID == palmDeviceID_TE 		||
	   deviceID == palmDeviceID_Zire21	||
	   deviceID == palmDeviceID_Zire71	|| // 71 has other byte order!
	   deviceID == palmDeviceID_Treo600 )
	   hasOMAPCPU = true;
	   
	return hasOMAPCPU;
}

Boolean DeviceHasMXLCPU()
{
	UInt32 deviceID=0;
	Boolean hasMXLCPU=false;
	
	// Should never fail
	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);
	
	
	if(deviceID == sonyHwrOEMDeviceID_TJ25 ||
	   deviceID == sonyHwrOEMDeviceID_TJ35 ||
	   deviceID == palmDeviceID_iQ3600)
	   //deviceID == palmDeviceID_Zodiac	)
	   hasMXLCPU = true;
	   
	return hasMXLCPU;
}
